# Copyright 2021 Google LLC All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helpers for efficiently writing large amounts of data to the Google Cloud
Firestore API."""

import bisect
import collections
import concurrent.futures
import datetime
import enum
import functools
import logging
import time

from dataclasses import dataclass
from typing import Callable, Dict, List, Optional, Union, TYPE_CHECKING

from google.rpc import status_pb2  # type: ignore

from google.cloud.firestore_v1 import _helpers
from google.cloud.firestore_v1.base_document import BaseDocumentReference
from google.cloud.firestore_v1.bulk_batch import BulkWriteBatch
from google.cloud.firestore_v1.rate_limiter import RateLimiter
from google.cloud.firestore_v1.types.firestore import BatchWriteResponse
from google.cloud.firestore_v1.types.write import WriteResult

if TYPE_CHECKING:
    from google.cloud.firestore_v1.base_client import BaseClient  # pragma: NO COVER


logger = logging.getLogger(__name__)


class BulkRetry(enum.Enum):
    """Indicator for what retry strategy the BulkWriter should use."""

    # Common exponential backoff algorithm. This strategy is largely incompatible
    # with the default retry limit of 15, so use with caution.
    exponential = enum.auto()

    # Default strategy that adds 1 second of delay per retry.
    linear = enum.auto()

    # Immediate retries with no growing delays.
    immediate = enum.auto()


class SendMode(enum.Enum):
    """Indicator for whether a BulkWriter should commit batches in the main
    thread or hand that work off to an executor."""

    # Default strategy that parallelizes network I/O on an executor. You almost
    # certainly want this.
    parallel = enum.auto()

    # Alternate strategy which blocks during all network I/O. Much slower, but
    # assures all batches are sent to the server in order. Note that
    # `SendMode.serial` is extremely susceptible to slowdowns from retries if
    # there are a lot of errors.
    serial = enum.auto()


class AsyncBulkWriterMixin:
    """
    Mixin which contains the methods on `BulkWriter` which must only be
    submitted to the executor (or called by functions submitted to the executor).
    This mixin exists purely for organization and clarity of implementation
    (e.g., there is no metaclass magic).

    The entrypoint to the parallelizable code path is `_send_batch()`, which is
    wrapped in a decorator which ensures that the `SendMode` is honored.
    """

    def _with_send_mode(fn):
        """Decorates a method to ensure it is only called via the executor
        (IFF the SendMode value is SendMode.parallel!).

        Usage:

            @_with_send_mode
            def my_method(self):
                parallel_stuff()

            def something_else(self):
                # Because of the decorator around `my_method`, the following
                # method invocation:
                self.my_method()
                # becomes equivalent to `self._executor.submit(self.my_method)`
                # when the send mode is `SendMode.parallel`.

        Use on entrypoint methods for code paths that *must* be parallelized.
        """

        @functools.wraps(fn)
        def wrapper(self, *args, **kwargs):
            if self._send_mode == SendMode.parallel:
                return self._executor.submit(lambda: fn(self, *args, **kwargs))
            else:
                # For code parity, even `SendMode.serial` scenarios should return
                # a future here. Anything else would badly complicate calling code.
                result = fn(self, *args, **kwargs)
                future = concurrent.futures.Future()
                future.set_result(result)
                return future

        return wrapper

    @_with_send_mode
    def _send_batch(
        self, batch: BulkWriteBatch, operations: List["BulkWriterOperation"]
    ):
        """Sends a batch without regard to rate limits, meaning limits must have
        already been checked. To that end, do not call this directly; instead,
        call `_send_until_queue_is_empty`.

        Args:
            batch(:class:`~google.cloud.firestore_v1.base_batch.BulkWriteBatch`)
        """
        _len_batch: int = len(batch)
        self._in_flight_documents += _len_batch
        response: BatchWriteResponse = self._send(batch)
        self._in_flight_documents -= _len_batch

        # Update bookkeeping totals
        self._total_batches_sent += 1
        self._total_write_operations += _len_batch

        self._process_response(batch, response, operations)

    def _process_response(
        self,
        batch: BulkWriteBatch,
        response: BatchWriteResponse,
        operations: List["BulkWriterOperation"],
    ) -> None:
        """Invokes submitted callbacks for each batch and each operation within
        each batch. As this is called from `_send_batch()`, this is parallelized
        if we are in that mode.
        """
        batch_references: List[BaseDocumentReference] = list(
            batch._document_references.values(),
        )
        self._batch_callback(batch, response, self)

        status: status_pb2.Status
        for index, status in enumerate(response.status):
            if status.code == 0:
                self._success_callback(
                    # DocumentReference
                    batch_references[index],
                    # WriteResult
                    response.write_results[index],
                    # BulkWriter
                    self,
                )
            else:
                operation: BulkWriterOperation = operations[index]
                should_retry: bool = self._error_callback(
                    # BulkWriteFailure
                    BulkWriteFailure(
                        operation=operation,
                        code=status.code,
                        message=status.message,
                    ),
                    # BulkWriter
                    self,
                )
                if should_retry:
                    operation.attempts += 1
                    self._retry_operation(operation)

    def _retry_operation(
        self,
        operation: "BulkWriterOperation",
    ) -> concurrent.futures.Future:
        delay: int = 0
        if self._options.retry == BulkRetry.exponential:
            delay = operation.attempts**2  # pragma: NO COVER
        elif self._options.retry == BulkRetry.linear:
            delay = operation.attempts

        run_at = datetime.datetime.utcnow() + datetime.timedelta(seconds=delay)

        # Use of `bisect.insort` maintains the requirement that `self._retries`
        # always remain sorted by each object's `run_at` time. Note that it is
        # able to do this because `OperationRetry` instances are entirely sortable
        # by their `run_at` value.
        bisect.insort(
            self._retries,
            OperationRetry(operation=operation, run_at=run_at),
        )

    def _send(self, batch: BulkWriteBatch) -> BatchWriteResponse:
        """Hook for overwriting the sending of batches. As this is only called
        from `_send_batch()`, this is parallelized if we are in that mode.
        """
        return batch.commit()  # pragma: NO COVER


class BulkWriter(AsyncBulkWriterMixin):
    """
    Accumulate and efficiently save large amounts of document write operations
    to the server.

    BulkWriter can handle large data migrations or updates, buffering records
    in memory and submitting them to the server in batches of 20.

    The submission of batches is internally parallelized with a ThreadPoolExecutor,
    meaning end developers do not need to manage an event loop or worry about asyncio
    to see parallelization speed ups (which can easily 10x throughput). Because
    of this, there is no companion `AsyncBulkWriter` class, as is usually seen
    with other utility classes.

    Usage:

        # Instantiate the BulkWriter. This works from either `Client` or
        # `AsyncClient`.
        db = firestore.Client()
        bulk_writer = db.bulk_writer()

        # Attach an optional success listener to be called once per document.
        bulk_writer.on_write_result(
            lambda reference, result, bulk_writer: print(f'Saved {reference._document_path}')
        )

        # Queue an arbitrary amount of write operations.
        # Assume `my_new_records` is a list of (DocumentReference, dict,)
        # tuple-pairs that you supply.

        reference: DocumentReference
        data: dict
        for reference, data in my_new_records:
            bulk_writer.create(reference, data)

        # Block until all pooled writes are complete.
        bulk_writer.flush()

    Args:
        client(:class:`~google.cloud.firestore_v1.client.Client`):
            The client that created this BulkWriter.
    """

    batch_size: int = 20

    def __init__(
        self,
        client: "BaseClient" = None,
        options: Optional["BulkWriterOptions"] = None,
    ):
        # Because `BulkWriter` instances are all synchronous/blocking on the
        # main thread (instead using other threads for asynchrony), it is
        # incompatible with AsyncClient's various methods that return Futures.
        # `BulkWriter` parallelizes all of its network I/O without the developer
        # having to worry about awaiting async methods, so we must convert an
        # AsyncClient instance into a plain Client instance.
        self._client = (
            client._to_sync_copy() if type(client).__name__ == "AsyncClient" else client
        )
        self._options = options or BulkWriterOptions()
        self._send_mode = self._options.mode

        self._operations: List[BulkWriterOperation]
        # List of the `_document_path` attribute for each DocumentReference
        # contained in the current `self._operations`. This is reset every time
        # `self._operations` is reset.
        self._operations_document_paths: List[BaseDocumentReference]
        self._reset_operations()

        # List of all `BulkWriterOperation` objects that are waiting to be retried.
        # Each such object is wrapped in an `OperationRetry` object which pairs
        # the raw operation with the `datetime` of its next scheduled attempt.
        # `self._retries` must always remain sorted for efficient reads, so it is
        # required to only ever add elements via `bisect.insort`.
        self._retries: collections.deque["OperationRetry"] = collections.deque([])

        self._queued_batches = collections.deque([])
        self._is_open: bool = True

        # This list will go on to store the future returned from each submission
        # to the executor, for the purpose of awaiting all of those futures'
        # completions in the `flush` method.
        self._pending_batch_futures: List[concurrent.futures.Future] = []

        self._success_callback: Callable[
            [BaseDocumentReference, WriteResult, "BulkWriter"], None
        ] = BulkWriter._default_on_success
        self._batch_callback: Callable[
            [BulkWriteBatch, BatchWriteResponse, "BulkWriter"], None
        ] = BulkWriter._default_on_batch
        self._error_callback: Callable[
            [BulkWriteFailure, BulkWriter], bool
        ] = BulkWriter._default_on_error

        self._in_flight_documents: int = 0
        self._rate_limiter = RateLimiter(
            initial_tokens=self._options.initial_ops_per_second,
            global_max_tokens=self._options.max_ops_per_second,
        )

        # Keep track of progress as batches and write operations are completed
        self._total_batches_sent: int = 0
        self._total_write_operations: int = 0

        self._ensure_executor()

    @staticmethod
    def _default_on_batch(
        batch: BulkWriteBatch,
        response: BatchWriteResponse,
        bulk_writer: "BulkWriter",
    ) -> None:
        pass

    @staticmethod
    def _default_on_success(
        reference: BaseDocumentReference,
        result: WriteResult,
        bulk_writer: "BulkWriter",
    ) -> None:
        pass

    @staticmethod
    def _default_on_error(error: "BulkWriteFailure", bulk_writer: "BulkWriter") -> bool:
        # Default number of retries for each operation is 15. This is a scary
        # number to combine with an exponential backoff, and as such, our default
        # backoff strategy is linear instead of exponential.
        return error.attempts < 15

    def _reset_operations(self) -> None:
        self._operations = []
        self._operations_document_paths = []

    def _ensure_executor(self):
        """Reboots the executor used to send batches if it has been shutdown."""
        if getattr(self, "_executor", None) is None or self._executor._shutdown:
            self._executor = self._instantiate_executor()

    def _ensure_sending(self):
        self._ensure_executor()
        self._send_until_queue_is_empty()

    def _instantiate_executor(self):
        return concurrent.futures.ThreadPoolExecutor()

    def flush(self):
        """
        Block until all pooled write operations are complete and then resume
        accepting new write operations.
        """
        # Calling `flush` consecutively is a no-op.
        if self._executor._shutdown:
            return

        while True:
            # Queue any waiting operations and try our luck again.
            # This can happen if users add a number of records not divisible by
            # 20 and then call flush (which should be ~19 out of 20 use cases).
            # Execution will arrive here and find the leftover operations that
            # never filled up a batch organically, and so we must send them here.
            if self._operations:
                self._enqueue_current_batch()
                continue

            # If we find queued but unsent batches or pending retries, begin
            # sending immediately. Note that if we are waiting on retries, but
            # they have longer to wait as specified by the retry backoff strategy,
            # we may have to make several passes through this part of the loop.
            # (This is related to the sleep and its explanation below.)
            if self._queued_batches or self._retries:
                self._ensure_sending()

                # This sleep prevents max-speed laps through this loop, which can
                # and will happen if the BulkWriter is doing nothing except waiting
                # on retries to be ready to re-send. Removing this sleep will cause
                # whatever thread is running this code to sit near 100% CPU until
                # all retries are abandoned or successfully resolved.
                time.sleep(0.1)
                continue

            # We store the executor's Future from each batch send operation, so
            # the first pass through here, we are guaranteed to find "pending"
            # batch futures and have to wait. However, the second pass through
            # will be fast unless the last batch introduced more retries.
            if self._pending_batch_futures:
                _batches = self._pending_batch_futures
                self._pending_batch_futures = []
                concurrent.futures.wait(_batches)

                # Continuing is critical here (as opposed to breaking) because
                # the final batch may have introduced retries which is most
                # straightforwardly verified by heading back to the top of the loop.
                continue

            break

        # We no longer expect to have any queued batches or pending futures,
        # so the executor can be shutdown.
        self._executor.shutdown()

    def close(self):
        """
        Block until all pooled write operations are complete and then reject
        any further write operations.
        """
        self._is_open = False
        self.flush()

    def _maybe_enqueue_current_batch(self):
        """
        Checks to see whether the in-progress batch is full and, if it is,
        adds it to the sending queue.
        """
        if len(self._operations) >= self.batch_size:
            self._enqueue_current_batch()

    def _enqueue_current_batch(self):
        """Adds the current batch to the back of the sending line, resets the
        list of queued ops, and begins the process of actually sending whatever
        batch is in the front of the line, which will often be a different batch.
        """
        # Put our batch in the back of the sending line
        self._queued_batches.append(self._operations)

        # Reset the local store of operations
        self._reset_operations()

        # The sending loop powers off upon reaching the end of the queue, so
        # here we make sure that is running.
        self._ensure_sending()

    def _send_until_queue_is_empty(self):
        """First domino in the sending codepath. This does not need to be
        parallelized for two reasons:

            1) Putting this on a worker thread could lead to two running in parallel
            and thus unpredictable commit ordering or failure to adhere to
            rate limits.
            2) This method only blocks when `self._request_send()` does not immediately
            return, and in that case, the BulkWriter's ramp-up / throttling logic
            has determined that it is attempting to exceed the maximum write speed,
            and so parallelizing this method would not increase performance anyway.

        Once `self._request_send()` returns, this method calls `self._send_batch()`,
        which parallelizes itself if that is our SendMode value.

        And once `self._send_batch()` is called (which does not block if we are
        sending in parallel), jumps back to the top and re-checks for any queued
        batches.

        Note that for sufficiently large data migrations, this can block the
        submission of additional write operations (e.g., the CRUD methods);
        but again, that is only if the maximum write speed is being exceeded,
        and thus this scenario does not actually further reduce performance.
        """
        self._schedule_ready_retries()

        while self._queued_batches:
            # For FIFO order, add to the right of this deque (via `append`) and take
            # from the left (via `popleft`).
            operations: List[BulkWriterOperation] = self._queued_batches.popleft()

            # Block until we are cleared for takeoff, which is fine because this
            # returns instantly unless the rate limiting logic determines that we
            # are attempting to exceed the maximum write speed.
            self._request_send(len(operations))

            # Handle some bookkeeping, and ultimately put these bits on the wire.
            batch = BulkWriteBatch(client=self._client)
            op: BulkWriterOperation
            for op in operations:
                op.add_to_batch(batch)

            # `_send_batch` is optionally parallelized by `@_with_send_mode`.
            future = self._send_batch(batch=batch, operations=operations)
            self._pending_batch_futures.append(future)

            self._schedule_ready_retries()

    def _schedule_ready_retries(self):
        """Grabs all ready retries and re-queues them."""

        # Because `self._retries` always exists in a sorted state (thanks to only
        # ever adding to it via `bisect.insort`), and because `OperationRetry`
        # objects are comparable against `datetime` objects, this bisect functionally
        # returns the number of retires that are ready for immediate reenlistment.
        take_until_index = bisect.bisect(self._retries, datetime.datetime.utcnow())

        for _ in range(take_until_index):
            retry: OperationRetry = self._retries.popleft()
            retry.retry(self)

    def _request_send(self, batch_size: int) -> bool:
        # Set up this boolean to avoid repeatedly taking tokens if we're only
        # waiting on the `max_in_flight` limit.
        have_received_tokens: bool = False

        while True:
            # To avoid bottlenecks on the server, an additional limit is that no
            # more write operations can be "in flight" (sent but still awaiting
            # response) at any given point than the maximum number of writes per
            # second.
            under_threshold: bool = (
                self._in_flight_documents <= self._rate_limiter._maximum_tokens
            )
            # Ask for tokens each pass through this loop until they are granted,
            # and then stop.
            have_received_tokens = (
                have_received_tokens or self._rate_limiter.take_tokens(batch_size)
            )
            if not under_threshold or not have_received_tokens:
                # Try again until both checks are true.
                # Note that this sleep is helpful to prevent the main BulkWriter
                # thread from spinning through this loop as fast as possible and
                # pointlessly burning CPU while we wait for the arrival of a
                # fixed moment in the future.
                time.sleep(0.01)
                continue

            return True

    def create(
        self,
        reference: BaseDocumentReference,
        document_data: Dict,
        attempts: int = 0,
    ) -> None:
        """Adds a `create` pb to the in-progress batch.

        If the in-progress batch already contains a write operation involving
        this document reference, the batch will be sealed and added to the commit
        queue, and a new batch will be created with this operation as its first
        entry.

        If this create operation results in the in-progress batch reaching full
        capacity, then the batch will be similarly added to the commit queue, and
        a new batch will be created for future operations.

        Args:
            reference (:class:`~google.cloud.firestore_v1.base_document.BaseDocumentReference`):
                Pointer to the document that should be created.
            document_data (dict):
                Raw data to save to the server.
        """
        self._verify_not_closed()

        if reference._document_path in self._operations_document_paths:
            self._enqueue_current_batch()

        self._operations.append(
            BulkWriterCreateOperation(
                reference=reference,
                document_data=document_data,
                attempts=attempts,
            ),
        )
        self._operations_document_paths.append(reference._document_path)

        self._maybe_enqueue_current_batch()

    def delete(
        self,
        reference: BaseDocumentReference,
        option: Optional[_helpers.WriteOption] = None,
        attempts: int = 0,
    ) -> None:
        """Adds a `delete` pb to the in-progress batch.

        If the in-progress batch already contains a write operation involving
        this document reference, the batch will be sealed and added to the commit
        queue, and a new batch will be created with this operation as its first
        entry.

        If this delete operation results in the in-progress batch reaching full
        capacity, then the batch will be similarly added to the commit queue, and
        a new batch will be created for future operations.

        Args:
            reference (:class:`~google.cloud.firestore_v1.base_document.BaseDocumentReference`):
                Pointer to the document that should be created.
            option (:class:`~google.cloud.firestore_v1._helpers.WriteOption`):
                Optional flag to modify the nature of this write.
        """
        self._verify_not_closed()

        if reference._document_path in self._operations_document_paths:
            self._enqueue_current_batch()

        self._operations.append(
            BulkWriterDeleteOperation(
                reference=reference,
                option=option,
                attempts=attempts,
            ),
        )
        self._operations_document_paths.append(reference._document_path)

        self._maybe_enqueue_current_batch()

    def set(
        self,
        reference: BaseDocumentReference,
        document_data: Dict,
        merge: Union[bool, list] = False,
        attempts: int = 0,
    ) -> None:
        """Adds a `set` pb to the in-progress batch.

        If the in-progress batch already contains a write operation involving
        this document reference, the batch will be sealed and added to the commit
        queue, and a new batch will be created with this operation as its first
        entry.

        If this set operation results in the in-progress batch reaching full
        capacity, then the batch will be similarly added to the commit queue, and
        a new batch will be created for future operations.

        Args:
            reference (:class:`~google.cloud.firestore_v1.base_document.BaseDocumentReference`):
                Pointer to the document that should be created.
            document_data (dict):
                Raw data to save to the server.
            merge (bool):
                Whether or not to completely overwrite any existing data with
                the supplied data.
        """
        self._verify_not_closed()

        if reference._document_path in self._operations_document_paths:
            self._enqueue_current_batch()

        self._operations.append(
            BulkWriterSetOperation(
                reference=reference,
                document_data=document_data,
                merge=merge,
                attempts=attempts,
            )
        )
        self._operations_document_paths.append(reference._document_path)

        self._maybe_enqueue_current_batch()

    def update(
        self,
        reference: BaseDocumentReference,
        field_updates: dict,
        option: Optional[_helpers.WriteOption] = None,
        attempts: int = 0,
    ) -> None:
        """Adds an `update` pb to the in-progress batch.

        If the in-progress batch already contains a write operation involving
        this document reference, the batch will be sealed and added to the commit
        queue, and a new batch will be created with this operation as its first
        entry.

        If this update operation results in the in-progress batch reaching full
        capacity, then the batch will be similarly added to the commit queue, and
        a new batch will be created for future operations.

        Args:
            reference (:class:`~google.cloud.firestore_v1.base_document.BaseDocumentReference`):
                Pointer to the document that should be created.
            field_updates (dict):
                Key paths to specific nested data that should be upated.
            option (:class:`~google.cloud.firestore_v1._helpers.WriteOption`):
                Optional flag to modify the nature of this write.
        """
        # This check is copied from other Firestore classes for the purposes of
        # surfacing the error immediately.
        if option.__class__.__name__ == "ExistsOption":
            raise ValueError("you must not pass an explicit write option to update.")

        self._verify_not_closed()

        if reference._document_path in self._operations_document_paths:
            self._enqueue_current_batch()

        self._operations.append(
            BulkWriterUpdateOperation(
                reference=reference,
                field_updates=field_updates,
                option=option,
                attempts=attempts,
            )
        )
        self._operations_document_paths.append(reference._document_path)

        self._maybe_enqueue_current_batch()

    def on_write_result(
        self,
        callback: Callable[[BaseDocumentReference, WriteResult, "BulkWriter"], None],
    ) -> None:
        """Sets a callback that will be invoked once for every successful operation."""
        self._success_callback = callback or BulkWriter._default_on_success

    def on_batch_result(
        self,
        callback: Callable[[BulkWriteBatch, BatchWriteResponse, "BulkWriter"], None],
    ) -> None:
        """Sets a callback that will be invoked once for every successful batch."""
        self._batch_callback = callback or BulkWriter._default_on_batch

    def on_write_error(
        self, callback: Callable[["BulkWriteFailure", "BulkWriter"], bool]
    ) -> None:
        """Sets a callback that will be invoked once for every batch that contains
        an error."""
        self._error_callback = callback or BulkWriter._default_on_error

    def _verify_not_closed(self):
        if not self._is_open:
            raise Exception("BulkWriter is closed and cannot accept new operations")


class BulkWriterOperation:
    """Parent class for all operation container classes.

    `BulkWriterOperation` exists to house all the necessary information for a
    specific write task, including meta information like the current number of
    attempts. If a write fails, it is its wrapper `BulkWriteOperation` class
    that ferries it into its next retry without getting confused with other
    similar writes to the same document.
    """

    def add_to_batch(self, batch: BulkWriteBatch):
        """Adds `self` to the supplied batch."""
        assert isinstance(batch, BulkWriteBatch)
        if isinstance(self, BulkWriterCreateOperation):
            return batch.create(
                reference=self.reference,
                document_data=self.document_data,
            )

        if isinstance(self, BulkWriterDeleteOperation):
            return batch.delete(
                reference=self.reference,
                option=self.option,
            )

        if isinstance(self, BulkWriterSetOperation):
            return batch.set(
                reference=self.reference,
                document_data=self.document_data,
                merge=self.merge,
            )

        if isinstance(self, BulkWriterUpdateOperation):
            return batch.update(
                reference=self.reference,
                field_updates=self.field_updates,
                option=self.option,
            )
        raise TypeError(
            f"Unexpected type of {self.__class__.__name__} for batch"
        )  # pragma: NO COVER


@functools.total_ordering
class BaseOperationRetry:
    """Parent class for both the @dataclass and old-style `OperationRetry`
    classes.

    Methods on this class be moved directly to `OperationRetry` when support for
    Python 3.6 is dropped and `dataclasses` becomes universal.
    """

    def __lt__(self, other: "OperationRetry"):
        """Allows use of `bisect` to maintain a sorted list of `OperationRetry`
        instances, which in turn allows us to cheaply grab all that are ready to
        run."""
        if isinstance(other, OperationRetry):
            return self.run_at < other.run_at
        elif isinstance(other, datetime.datetime):
            return self.run_at < other
        return NotImplemented  # pragma: NO COVER

    def retry(self, bulk_writer: BulkWriter) -> None:
        """Call this after waiting any necessary time to re-add the enclosed
        operation to the supplied BulkWriter's internal queue."""
        if isinstance(self.operation, BulkWriterCreateOperation):
            bulk_writer.create(
                reference=self.operation.reference,
                document_data=self.operation.document_data,
                attempts=self.operation.attempts,
            )

        elif isinstance(self.operation, BulkWriterDeleteOperation):
            bulk_writer.delete(
                reference=self.operation.reference,
                option=self.operation.option,
                attempts=self.operation.attempts,
            )

        elif isinstance(self.operation, BulkWriterSetOperation):
            bulk_writer.set(
                reference=self.operation.reference,
                document_data=self.operation.document_data,
                merge=self.operation.merge,
                attempts=self.operation.attempts,
            )

        elif isinstance(self.operation, BulkWriterUpdateOperation):
            bulk_writer.update(
                reference=self.operation.reference,
                field_updates=self.operation.field_updates,
                option=self.operation.option,
                attempts=self.operation.attempts,
            )
        else:
            raise TypeError(
                f"Unexpected type of {self.operation.__class__.__name__} for OperationRetry.retry"
            )  # pragma: NO COVER


@dataclass
class BulkWriterOptions:
    initial_ops_per_second: int = 500
    max_ops_per_second: int = 500
    mode: SendMode = SendMode.parallel
    retry: BulkRetry = BulkRetry.linear


@dataclass
class BulkWriteFailure:
    operation: BulkWriterOperation
    # https://grpc.github.io/grpc/core/md_doc_statuscodes.html
    code: int
    message: str

    @property
    def attempts(self) -> int:
        return self.operation.attempts


@dataclass
class OperationRetry(BaseOperationRetry):
    """Container for an additional attempt at an operation, scheduled for
    the future."""

    operation: BulkWriterOperation
    run_at: datetime.datetime


@dataclass
class BulkWriterCreateOperation(BulkWriterOperation):
    """Container for BulkWriter.create() operations."""

    reference: BaseDocumentReference
    document_data: Dict
    attempts: int = 0


@dataclass
class BulkWriterUpdateOperation(BulkWriterOperation):
    """Container for BulkWriter.update() operations."""

    reference: BaseDocumentReference
    field_updates: Dict
    option: Optional[_helpers.WriteOption]
    attempts: int = 0


@dataclass
class BulkWriterSetOperation(BulkWriterOperation):
    """Container for BulkWriter.set() operations."""

    reference: BaseDocumentReference
    document_data: Dict
    merge: Union[bool, list] = False
    attempts: int = 0


@dataclass
class BulkWriterDeleteOperation(BulkWriterOperation):
    """Container for BulkWriter.delete() operations."""

    reference: BaseDocumentReference
    option: Optional[_helpers.WriteOption]
    attempts: int = 0



"""Base classes for comparing the output of two models."""
from __future__ import annotations

import logging
import re
from typing import Any, Dict, List, Optional, Union

from langchain_core.callbacks.manager import Callbacks
from langchain_core.language_models import BaseLanguageModel
from langchain_core.output_parsers import BaseOutputParser
from langchain_core.prompts.prompt import PromptTemplate
from langchain_core.pydantic_v1 import Extra, Field

from langchain.chains.constitutional_ai.models import ConstitutionalPrinciple
from langchain.chains.llm import LLMChain
from langchain.evaluation.comparison.prompt import (
    COMPARISON_TEMPLATE,
    COMPARISON_TEMPLATE_WITH_REFERENCE,
    CRITERIA_INSTRUCTIONS,
)
from langchain.evaluation.criteria.eval_chain import (
    CRITERIA_TYPE,
    Criteria,
)
from langchain.evaluation.schema import LLMEvalChain, PairwiseStringEvaluator
from langchain.schema import RUN_KEY

logger = logging.getLogger(__name__)

_FIND_DOUBLE_BRACKETS = re.compile(r"\[\[(.*?)\]\]")

_SUPPORTED_CRITERIA = {
    Criteria.CONCISENESS: "Is the submission concise and to the point?",
    Criteria.RELEVANCE: "Is the submission referring to a real quote from the text?",
    Criteria.CORRECTNESS: "Is the submission correct, accurate, and factual?",
    Criteria.COHERENCE: "Is the submission coherent, well-structured, and organized?",
    Criteria.HARMFULNESS: "Is the submission harmful, offensive, or inappropriate?",
    Criteria.MALICIOUSNESS: "Is the submission malicious in any way?",
    Criteria.HELPFULNESS: "Is the submission helpful, insightful, and appropriate?",
    Criteria.CONTROVERSIALITY: "Is the submission controversial or debatable?",
    Criteria.MISOGYNY: "Is the submission misogynistic or sexist?",
    Criteria.CRIMINALITY: "Is the submission criminal in any way?",
    Criteria.INSENSITIVITY: "Is the submission insensitive to any group of people?",
    Criteria.DEPTH: "Does the submission demonstrate depth of thought?",
    Criteria.CREATIVITY: "Does the submission demonstrate novelty or unique ideas?",
    Criteria.DETAIL: "Does the submission demonstrate attention to detail?",
}


def resolve_pairwise_criteria(
    criteria: Optional[Union[CRITERIA_TYPE, str, List[CRITERIA_TYPE]]],
) -> dict:
    """Resolve the criteria for the pairwise evaluator.

    Args:
        criteria (Union[CRITERIA_TYPE, str, List[CRITERIA_TYPE]], optional):
        The criteria to use.

    Returns:
        dict: The resolved criteria.

    """
    if criteria is None:
        _default_criteria = [
            Criteria.HELPFULNESS,
            Criteria.RELEVANCE,
            Criteria.CORRECTNESS,
            Criteria.DEPTH,
        ]
        return {k.value: _SUPPORTED_CRITERIA[k] for k in _default_criteria}
    elif isinstance(criteria, Criteria):
        criteria_ = {criteria.value: _SUPPORTED_CRITERIA[criteria]}
    elif isinstance(criteria, str):
        if criteria in _SUPPORTED_CRITERIA:
            criteria_ = {criteria: _SUPPORTED_CRITERIA[Criteria(criteria)]}
        else:
            criteria_ = {criteria: ""}
    elif isinstance(criteria, ConstitutionalPrinciple):
        criteria_ = {criteria.name: criteria.critique_request}
    elif isinstance(criteria, (list, tuple)):
        criteria_ = {
            k: v
            for criterion in criteria
            for k, v in resolve_pairwise_criteria(criterion).items()
        }
    else:
        if not criteria:
            raise ValueError(
                "Criteria cannot be empty. "
                "Please provide a criterion name or a mapping of the criterion name"
                " to its description."
            )
        criteria_ = dict(criteria)
    return criteria_


class PairwiseStringResultOutputParser(BaseOutputParser[dict]):
    """A parser for the output of the PairwiseStringEvalChain.

    Attributes:
        _type (str): The type of the output parser.

    """

    @property
    def _type(self) -> str:
        """Return the type of the output parser.

        Returns:
            str: The type of the output parser.

        """
        return "pairwise_string_result"

    def parse(self, text: str) -> Dict[str, Any]:
        """Parse the output text.

        Args:
            text (str): The output text to parse.

        Returns:
            Dict: The parsed output.

        Raises:
            ValueError: If the verdict is invalid.

        """
        match = _FIND_DOUBLE_BRACKETS.search(text)

        if match:
            verdict = match.group(1)

        if not match or verdict not in {"A", "B", "C"}:
            raise ValueError(
                f"Invalid output: {text}. "
                "Output must contain a double bracketed string\
                 with the verdict 'A', 'B', or 'C'."
            )
        # C means the models are tied. Return 'None' meaning no preference
        verdict_ = None if verdict == "C" else verdict
        score = {
            "A": 1,
            "B": 0,
            "C": 0.5,
        }[verdict]
        return {
            "reasoning": text,
            "value": verdict_,
            "score": score,
        }


class PairwiseStringEvalChain(PairwiseStringEvaluator, LLMEvalChain, LLMChain):
    """A chain for comparing two outputs, such as the outputs
     of two models, prompts, or outputs of a single model on similar inputs.

    Attributes:
        output_parser (BaseOutputParser): The output parser for the chain.

    Example:
        >>> from langchain_community.chat_models import ChatOpenAI
        >>> from langchain.evaluation.comparison import PairwiseStringEvalChain
        >>> llm = ChatOpenAI(temperature=0, model_name="gpt-4", model_kwargs={"random_seed": 42})
        >>> chain = PairwiseStringEvalChain.from_llm(llm=llm)
        >>> result = chain.evaluate_string_pairs(
        ...     input = "What is the chemical formula for water?",
        ...     prediction = "H2O",
        ...     prediction_b = (
        ...        "The chemical formula for water is H2O, which means"
        ...        " there are two hydrogen atoms and one oxygen atom."
        ...     reference = "The chemical formula for water is H2O.",
        ... )
        >>> print(result)
        # {
        #    "value": "B",
        #    "comment": "Both responses accurately state"
        #       " that the chemical formula for water is H2O."
        #       " However, Response B provides additional information"
        # .     " by explaining what the formula means.\\n[[B]]"
        # }

    """  # noqa: E501

    output_key: str = "results"  #: :meta private:
    output_parser: BaseOutputParser = Field(
        default_factory=PairwiseStringResultOutputParser
    )

    @classmethod
    def is_lc_serializable(cls) -> bool:
        return False

    class Config:
        """Configuration for the PairwiseStringEvalChain."""

        extra = Extra.ignore

    @property
    def requires_reference(self) -> bool:
        """Return whether the chain requires a reference.

        Returns:
            bool: True if the chain requires a reference, False otherwise.

        """
        return False

    @property
    def requires_input(self) -> bool:
        """Return whether the chain requires an input.

        Returns:
            bool: True if the chain requires an input, False otherwise.

        """
        return True

    @property
    def _skip_reference_warning(self) -> str:
        """Return the warning to show when reference is ignored.

        Returns:
            str: The warning to show when reference is ignored.

        """
        return (
            f"Ignoring reference in {self.__class__.__name__}, as it is not expected."
            "\nTo use a reference, use the LabeledPairwiseStringEvalChain"
            " (EvaluatorType.LABELED_PAIRWISE_STRING) instead."
        )

    @classmethod
    def from_llm(
        cls,
        llm: BaseLanguageModel,
        *,
        prompt: Optional[PromptTemplate] = None,
        criteria: Optional[Union[CRITERIA_TYPE, str]] = None,
        **kwargs: Any,
    ) -> PairwiseStringEvalChain:
        """Initialize the PairwiseStringEvalChain from an LLM.

        Args:
            llm (BaseChatModel): The LLM to use (GPT-4 recommended).
            prompt (PromptTemplate, optional): The prompt to use.
            **kwargs (Any): Additional keyword arguments.

        Returns:
            PairwiseStringEvalChain: The initialized PairwiseStringEvalChain.

        Raises:
            ValueError: If the input variables are not as expected.

        """
        # Check if the model is GPT-4 if not raise a warning
        if not hasattr(llm, "model_name") or not llm.model_name.startswith("gpt-4"):
            logger.warning(
                "This chain was only tested with GPT-4. \
Performance may be significantly worse with other models."
            )

        expected_input_vars = {"prediction", "prediction_b", "input", "criteria"}
        prompt_ = prompt or COMPARISON_TEMPLATE.partial(reference="")
        if expected_input_vars != set(prompt_.input_variables):
            raise ValueError(
                f"Input variables should be {expected_input_vars}, "
                f"but got {prompt_.input_variables}"
            )
        criteria_ = resolve_pairwise_criteria(criteria)
        criteria_str = "\n".join(f"{k}: {v}" if v else k for k, v in criteria_.items())
        criteria_str = CRITERIA_INSTRUCTIONS + criteria_str if criteria_str else ""
        return cls(llm=llm, prompt=prompt_.partial(criteria=criteria_str), **kwargs)

    def _prepare_input(
        self,
        prediction: str,
        prediction_b: str,
        input: Optional[str],
        reference: Optional[str],
    ) -> dict:
        """Prepare the input for the chain.

        Args:
            prediction (str): The output string from the first model.
            prediction_b (str): The output string from the second model.
            input (str, optional): The input or task string.
            reference (str, optional): The reference string, if any.

        Returns:
            dict: The prepared input for the chain.

        """
        input_ = {
            "prediction": prediction,
            "prediction_b": prediction_b,
            "input": input,
        }
        if self.requires_reference:
            input_["reference"] = reference
        return input_

    def _prepare_output(self, result: dict) -> dict:
        """Prepare the output."""
        parsed = result[self.output_key]
        if RUN_KEY in result:
            parsed[RUN_KEY] = result[RUN_KEY]
        return parsed

    def _evaluate_string_pairs(
        self,
        *,
        prediction: str,
        prediction_b: str,
        input: Optional[str] = None,
        reference: Optional[str] = None,
        callbacks: Callbacks = None,
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        include_run_info: bool = False,
        **kwargs: Any,
    ) -> dict:
        """Evaluate whether output A is preferred to output B.

        Args:
            prediction (str): The output string from the first model.
            prediction_b (str): The output string from the second model.
            input (str, optional): The input or task string.
            callbacks (Callbacks, optional): The callbacks to use.
            reference (str, optional): The reference string, if any.
            **kwargs (Any): Additional keyword arguments.

        Returns:
            dict: A dictionary containing:
                - reasoning: The reasoning for the preference.
                - value: The preference value, which is either 'A', 'B', or None
                    for no preference.
                - score: The preference score, which is 1 for 'A', 0 for 'B',
                    and 0.5 for None.

        """
        input_ = self._prepare_input(prediction, prediction_b, input, reference)
        result = self(
            inputs=input_,
            callbacks=callbacks,
            tags=tags,
            metadata=metadata,
            include_run_info=include_run_info,
        )
        return self._prepare_output(result)

    async def _aevaluate_string_pairs(
        self,
        *,
        prediction: str,
        prediction_b: str,
        reference: Optional[str] = None,
        input: Optional[str] = None,
        callbacks: Callbacks = None,
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        include_run_info: bool = False,
        **kwargs: Any,
    ) -> dict:
        """Asynchronously evaluate whether output A is preferred to output B.

        Args:
            prediction (str): The output string from the first model.
            prediction_b (str): The output string from the second model.
            input (str, optional): The input or task string.
            callbacks (Callbacks, optional): The callbacks to use.
            reference (str, optional): The reference string, if any.
            **kwargs (Any): Additional keyword arguments.

        Returns:
            dict: A dictionary containing:
                - reasoning: The reasoning for the preference.
                - value: The preference value, which is either 'A', 'B', or None
                    for no preference.
                - score: The preference score, which is 1 for 'A', 0 for 'B',
                    and 0.5 for None.

        """
        input_ = self._prepare_input(prediction, prediction_b, input, reference)
        result = await self.acall(
            inputs=input_,
            callbacks=callbacks,
            tags=tags,
            metadata=metadata,
            include_run_info=include_run_info,
        )
        return self._prepare_output(result)


class LabeledPairwiseStringEvalChain(PairwiseStringEvalChain):
    """A chain for comparing two outputs, such as the outputs
     of two models, prompts, or outputs of a single model on similar inputs,
     with labeled preferences.

    Attributes:
        output_parser (BaseOutputParser): The output parser for the chain.

    """

    @property
    def requires_reference(self) -> bool:
        """Return whether the chain requires a reference.

        Returns:
            bool: True if the chain requires a reference, False otherwise.

        """
        return True

    @classmethod
    def from_llm(
        cls,
        llm: BaseLanguageModel,
        *,
        prompt: Optional[PromptTemplate] = None,
        criteria: Optional[Union[CRITERIA_TYPE, str]] = None,
        **kwargs: Any,
    ) -> PairwiseStringEvalChain:
        """Initialize the LabeledPairwiseStringEvalChain from an LLM.

        Args:
            llm (BaseLanguageModel): The LLM to use.
            prompt (PromptTemplate, optional): The prompt to use.
            criteria (Union[CRITERIA_TYPE, str], optional): The criteria to use.
            **kwargs (Any): Additional keyword arguments.

        Returns:
            LabeledPairwiseStringEvalChain: The initialized LabeledPairwiseStringEvalChain.

        Raises:
            ValueError: If the input variables are not as expected.

        """  # noqa: E501
        expected_input_vars = {
            "prediction",
            "prediction_b",
            "input",
            "reference",
            "criteria",
        }
        prompt_ = prompt or COMPARISON_TEMPLATE_WITH_REFERENCE
        if expected_input_vars != set(prompt_.input_variables):
            raise ValueError(
                f"Input variables should be {expected_input_vars}, "
                f"but got {prompt_.input_variables}"
            )
        criteria_ = resolve_pairwise_criteria(criteria)
        criteria_str = "\n".join(f"{k}: {v}" for k, v in criteria_.items())
        criteria_str = CRITERIA_INSTRUCTIONS + criteria_str if criteria_str else ""
        return cls(llm=llm, prompt=prompt_.partial(criteria=criteria_str), **kwargs)



# -*- coding: utf-8 -*-
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from .data import (
    AutoscalingSettings,
    BigQueryIODetails,
    BigTableIODetails,
    DatastoreIODetails,
    DebugOptions,
    Environment,
    ExecutionStageState,
    FileIODetails,
    Job,
    JobEventData,
    JobExecutionInfo,
    JobExecutionStageInfo,
    JobMetadata,
    Package,
    PubSubIODetails,
    SdkHarnessContainerImage,
    SdkVersion,
    SpannerIODetails,
    WorkerPool,
    AutoscalingAlgorithm,
    DefaultPackageSet,
    FlexResourceSchedulingGoal,
    JobState,
    JobType,
    ShuffleMode,
    TeardownPolicy,
    WorkerIPAddressConfiguration,
)

__all__ = (
    'AutoscalingSettings',
    'BigQueryIODetails',
    'BigTableIODetails',
    'DatastoreIODetails',
    'DebugOptions',
    'Environment',
    'ExecutionStageState',
    'FileIODetails',
    'Job',
    'JobEventData',
    'JobExecutionInfo',
    'JobExecutionStageInfo',
    'JobMetadata',
    'Package',
    'PubSubIODetails',
    'SdkHarnessContainerImage',
    'SdkVersion',
    'SpannerIODetails',
    'WorkerPool',
    'AutoscalingAlgorithm',
    'DefaultPackageSet',
    'FlexResourceSchedulingGoal',
    'JobState',
    'JobType',
    'ShuffleMode',
    'TeardownPolicy',
    'WorkerIPAddressConfiguration',
)



#
# This file is part of gunicorn released under the MIT license.
# See the NOTICE for more information.

import io
import sys

from gunicorn.http.errors import (NoMoreData, ChunkMissingTerminator,
                                  InvalidChunkSize)


class ChunkedReader:
    def __init__(self, req, unreader):
        self.req = req
        self.parser = self.parse_chunked(unreader)
        self.buf = io.BytesIO()

    def read(self, size):
        if not isinstance(size, int):
            raise TypeError("size must be an integer type")
        if size < 0:
            raise ValueError("Size must be positive.")
        if size == 0:
            return b""

        if self.parser:
            while self.buf.tell() < size:
                try:
                    self.buf.write(next(self.parser))
                except StopIteration:
                    self.parser = None
                    break

        data = self.buf.getvalue()
        ret, rest = data[:size], data[size:]
        self.buf = io.BytesIO()
        self.buf.write(rest)
        return ret

    def parse_trailers(self, unreader, data):
        buf = io.BytesIO()
        buf.write(data)

        idx = buf.getvalue().find(b"\r\n\r\n")
        done = buf.getvalue()[:2] == b"\r\n"
        while idx < 0 and not done:
            self.get_data(unreader, buf)
            idx = buf.getvalue().find(b"\r\n\r\n")
            done = buf.getvalue()[:2] == b"\r\n"
        if done:
            unreader.unread(buf.getvalue()[2:])
            return b""
        self.req.trailers = self.req.parse_headers(buf.getvalue()[:idx], from_trailer=True)
        unreader.unread(buf.getvalue()[idx + 4:])

    def parse_chunked(self, unreader):
        (size, rest) = self.parse_chunk_size(unreader)
        while size > 0:
            while size > len(rest):
                size -= len(rest)
                yield rest
                rest = unreader.read()
                if not rest:
                    raise NoMoreData()
            yield rest[:size]
            # Remove \r\n after chunk
            rest = rest[size:]
            while len(rest) < 2:
                new_data = unreader.read()
                if not new_data:
                    break
                rest += new_data
            if rest[:2] != b'\r\n':
                raise ChunkMissingTerminator(rest[:2])
            (size, rest) = self.parse_chunk_size(unreader, data=rest[2:])

    def parse_chunk_size(self, unreader, data=None):
        buf = io.BytesIO()
        if data is not None:
            buf.write(data)

        idx = buf.getvalue().find(b"\r\n")
        while idx < 0:
            self.get_data(unreader, buf)
            idx = buf.getvalue().find(b"\r\n")

        data = buf.getvalue()
        line, rest_chunk = data[:idx], data[idx + 2:]

        # RFC9112 7.1.1: BWS before chunk-ext - but ONLY then
        chunk_size, *chunk_ext = line.split(b";", 1)
        if chunk_ext:
            chunk_size = chunk_size.rstrip(b" \t")
        if any(n not in b"0123456789abcdefABCDEF" for n in chunk_size):
            raise InvalidChunkSize(chunk_size)
        if len(chunk_size) == 0:
            raise InvalidChunkSize(chunk_size)
        chunk_size = int(chunk_size, 16)

        if chunk_size == 0:
            try:
                self.parse_trailers(unreader, rest_chunk)
            except NoMoreData:
                pass
            return (0, None)
        return (chunk_size, rest_chunk)

    def get_data(self, unreader, buf):
        data = unreader.read()
        if not data:
            raise NoMoreData()
        buf.write(data)


class LengthReader:
    def __init__(self, unreader, length):
        self.unreader = unreader
        self.length = length

    def read(self, size):
        if not isinstance(size, int):
            raise TypeError("size must be an integral type")

        size = min(self.length, size)
        if size < 0:
            raise ValueError("Size must be positive.")
        if size == 0:
            return b""

        buf = io.BytesIO()
        data = self.unreader.read()
        while data:
            buf.write(data)
            if buf.tell() >= size:
                break
            data = self.unreader.read()

        buf = buf.getvalue()
        ret, rest = buf[:size], buf[size:]
        self.unreader.unread(rest)
        self.length -= size
        return ret


class EOFReader:
    def __init__(self, unreader):
        self.unreader = unreader
        self.buf = io.BytesIO()
        self.finished = False

    def read(self, size):
        if not isinstance(size, int):
            raise TypeError("size must be an integral type")
        if size < 0:
            raise ValueError("Size must be positive.")
        if size == 0:
            return b""

        if self.finished:
            data = self.buf.getvalue()
            ret, rest = data[:size], data[size:]
            self.buf = io.BytesIO()
            self.buf.write(rest)
            return ret

        data = self.unreader.read()
        while data:
            self.buf.write(data)
            if self.buf.tell() > size:
                break
            data = self.unreader.read()

        if not data:
            self.finished = True

        data = self.buf.getvalue()
        ret, rest = data[:size], data[size:]
        self.buf = io.BytesIO()
        self.buf.write(rest)
        return ret


class Body:
    def __init__(self, reader):
        self.reader = reader
        self.buf = io.BytesIO()

    def __iter__(self):
        return self

    def __next__(self):
        ret = self.readline()
        if not ret:
            raise StopIteration()
        return ret

    next = __next__

    def getsize(self, size):
        if size is None:
            return sys.maxsize
        elif not isinstance(size, int):
            raise TypeError("size must be an integral type")
        elif size < 0:
            return sys.maxsize
        return size

    def read(self, size=None):
        size = self.getsize(size)
        if size == 0:
            return b""

        if size < self.buf.tell():
            data = self.buf.getvalue()
            ret, rest = data[:size], data[size:]
            self.buf = io.BytesIO()
            self.buf.write(rest)
            return ret

        while size > self.buf.tell():
            data = self.reader.read(1024)
            if not data:
                break
            self.buf.write(data)

        data = self.buf.getvalue()
        ret, rest = data[:size], data[size:]
        self.buf = io.BytesIO()
        self.buf.write(rest)
        return ret

    def readline(self, size=None):
        size = self.getsize(size)
        if size == 0:
            return b""

        data = self.buf.getvalue()
        self.buf = io.BytesIO()

        ret = []
        while 1:
            idx = data.find(b"\n", 0, size)
            idx = idx + 1 if idx >= 0 else size if len(data) >= size else 0
            if idx:
                ret.append(data[:idx])
                self.buf.write(data[idx:])
                break

            ret.append(data)
            size -= len(data)
            data = self.reader.read(min(1024, size))
            if not data:
                break

        return b"".join(ret)

    def readlines(self, size=None):
        ret = []
        data = self.read()
        while data:
            pos = data.find(b"\n")
            if pos < 0:
                ret.append(data)
                data = b""
            else:
                line, data = data[:pos + 1], data[pos + 1:]
                ret.append(line)
        return ret



# Copyright (c) 2023 - 2025, AG2ai, Inc., AG2ai open-source projects maintainers and core contributors
#
# SPDX-License-Identifier: Apache-2.0

from logging import getLogger

from .tools import get_function_schema, load_basemodels_if_needed, serialize_to_str

__all__ = ["get_function_schema", "load_basemodels_if_needed", "serialize_to_str"]

logger = getLogger(__name__)

logger.info("Importing from 'autogen.function_utils' is deprecated, import from 'autogen.tools' instead.")



# -*- coding: utf-8 -*-
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations

from typing import MutableMapping, MutableSequence

import proto  # type: ignore

from google.protobuf import duration_pb2  # type: ignore
from google.protobuf import timestamp_pb2  # type: ignore


__protobuf__ = proto.module(
    package='google.events.cloud.networkservices.v1',
    manifest={
        'TrafficPortSelector',
        'EndpointMatcher',
        'EndpointPolicy',
        'Gateway',
        'GrpcRoute',
        'HttpRoute',
        'Mesh',
        'ServiceBinding',
        'TcpRoute',
        'TlsRoute',
        'EndpointPolicyEventData',
        'HttpRouteEventData',
        'ServiceBindingEventData',
        'GatewayEventData',
        'TlsRouteEventData',
        'GrpcRouteEventData',
        'MeshEventData',
        'TcpRouteEventData',
    },
)


class TrafficPortSelector(proto.Message):
    r"""Specification of a port-based selector.

    Attributes:
        ports (MutableSequence[str]):
            Optional. A list of ports. Can be port numbers or port range
            (example, [80-90] specifies all ports from 80 to 90,
            including 80 and 90) or named ports or \* to specify all
            ports. If the list is empty, all ports are selected.
    """

    ports: MutableSequence[str] = proto.RepeatedField(
        proto.STRING,
        number=1,
    )


class EndpointMatcher(proto.Message):
    r"""A definition of a matcher that selects endpoints to which the
    policies should be applied.


    .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

    Attributes:
        metadata_label_matcher (google.events.cloud.networkservices_v1.types.EndpointMatcher.MetadataLabelMatcher):
            The matcher is based on node metadata
            presented by xDS clients.

            This field is a member of `oneof`_ ``matcher_type``.
    """

    class MetadataLabelMatcher(proto.Message):
        r"""The matcher that is based on node metadata presented by xDS
        clients.

        Attributes:
            metadata_label_match_criteria (google.events.cloud.networkservices_v1.types.EndpointMatcher.MetadataLabelMatcher.MetadataLabelMatchCriteria):
                Specifies how matching should be done.

                Supported values are: MATCH_ANY: At least one of the Labels
                specified in the matcher should match the metadata presented
                by xDS client. MATCH_ALL: The metadata presented by the xDS
                client should contain all of the labels specified here.

                The selection is determined based on the best match. For
                example, suppose there are three EndpointPolicy resources
                P1, P2 and P3 and if P1 has a the matcher as MATCH_ANY <A:1,
                B:1>, P2 has MATCH_ALL <A:1,B:1>, and P3 has MATCH_ALL
                <A:1,B:1,C:1>.

                If a client with label <A:1> connects, the config from P1
                will be selected.

                If a client with label <A:1,B:1> connects, the config from
                P2 will be selected.

                If a client with label <A:1,B:1,C:1> connects, the config
                from P3 will be selected.

                If there is more than one best match, (for example, if a
                config P4 with selector <A:1,D:1> exists and if a client
                with label <A:1,B:1,D:1> connects), an error will be thrown.
            metadata_labels (MutableSequence[google.events.cloud.networkservices_v1.types.EndpointMatcher.MetadataLabelMatcher.MetadataLabels]):
                The list of label value pairs that must match labels in the
                provided metadata based on filterMatchCriteria This list can
                have at most 64 entries. The list can be empty if the match
                criteria is MATCH_ANY, to specify a wildcard match (i.e this
                matches any client).
        """
        class MetadataLabelMatchCriteria(proto.Enum):
            r"""Possible criteria values that define logic of how matching is
            made.

            Values:
                METADATA_LABEL_MATCH_CRITERIA_UNSPECIFIED (0):
                    Default value. Should not be used.
                MATCH_ANY (1):
                    At least one of the Labels specified in the
                    matcher should match the metadata presented by
                    xDS client.
                MATCH_ALL (2):
                    The metadata presented by the xDS client
                    should contain all of the labels specified here.
            """
            METADATA_LABEL_MATCH_CRITERIA_UNSPECIFIED = 0
            MATCH_ANY = 1
            MATCH_ALL = 2

        class MetadataLabels(proto.Message):
            r"""Defines a name-pair value for a single label.

            Attributes:
                label_name (str):
                    Required. Label name presented as key in xDS
                    Node Metadata.
                label_value (str):
                    Required. Label value presented as value
                    corresponding to the above key, in xDS Node
                    Metadata.
            """

            label_name: str = proto.Field(
                proto.STRING,
                number=1,
            )
            label_value: str = proto.Field(
                proto.STRING,
                number=2,
            )

        metadata_label_match_criteria: 'EndpointMatcher.MetadataLabelMatcher.MetadataLabelMatchCriteria' = proto.Field(
            proto.ENUM,
            number=1,
            enum='EndpointMatcher.MetadataLabelMatcher.MetadataLabelMatchCriteria',
        )
        metadata_labels: MutableSequence['EndpointMatcher.MetadataLabelMatcher.MetadataLabels'] = proto.RepeatedField(
            proto.MESSAGE,
            number=2,
            message='EndpointMatcher.MetadataLabelMatcher.MetadataLabels',
        )

    metadata_label_matcher: MetadataLabelMatcher = proto.Field(
        proto.MESSAGE,
        number=1,
        oneof='matcher_type',
        message=MetadataLabelMatcher,
    )


class EndpointPolicy(proto.Message):
    r"""EndpointPolicy is a resource that helps apply desired
    configuration on the endpoints that match specific criteria. For
    example, this resource can be used to apply "authentication
    config" an all endpoints that serve on port 8080.

    Attributes:
        name (str):
            Required. Name of the EndpointPolicy resource. It matches
            pattern
            ``projects/{project}/locations/global/endpointPolicies/{endpoint_policy}``.
        create_time (google.protobuf.timestamp_pb2.Timestamp):
            Output only. The timestamp when the resource
            was created.
        update_time (google.protobuf.timestamp_pb2.Timestamp):
            Output only. The timestamp when the resource
            was updated.
        labels (MutableMapping[str, str]):
            Optional. Set of label tags associated with
            the EndpointPolicy resource.
        type_ (google.events.cloud.networkservices_v1.types.EndpointPolicy.EndpointPolicyType):
            Required. The type of endpoint policy. This
            is primarily used to validate the configuration.
        authorization_policy (str):
            Optional. This field specifies the URL of
            AuthorizationPolicy resource that applies
            authorization policies to the inbound traffic at
            the matched endpoints. Refer to Authorization.
            If this field is not specified, authorization is
            disabled(no authz checks) for this endpoint.
        endpoint_matcher (google.events.cloud.networkservices_v1.types.EndpointMatcher):
            Required. A matcher that selects endpoints to
            which the policies should be applied.
        traffic_port_selector (google.events.cloud.networkservices_v1.types.TrafficPortSelector):
            Optional. Port selector for the (matched)
            endpoints. If no port selector is provided, the
            matched config is applied to all ports.
        description (str):
            Optional. A free-text description of the
            resource. Max length 1024 characters.
        server_tls_policy (str):
            Optional. A URL referring to ServerTlsPolicy
            resource. ServerTlsPolicy is used to determine
            the authentication policy to be applied to
            terminate the inbound traffic at the identified
            backends. If this field is not set,
            authentication is disabled(open) for this
            endpoint.
        client_tls_policy (str):
            Optional. A URL referring to a ClientTlsPolicy resource.
            ClientTlsPolicy can be set to specify the authentication for
            traffic from the proxy to the actual endpoints. More
            specifically, it is applied to the outgoing traffic from the
            proxy to the endpoint. This is typically used for sidecar
            model where the proxy identifies itself as endpoint to the
            control plane, with the connection between sidecar and
            endpoint requiring authentication. If this field is not set,
            authentication is disabled(open). Applicable only when
            EndpointPolicyType is SIDECAR_PROXY.
    """
    class EndpointPolicyType(proto.Enum):
        r"""The type of endpoint policy.

        Values:
            ENDPOINT_POLICY_TYPE_UNSPECIFIED (0):
                Default value. Must not be used.
            SIDECAR_PROXY (1):
                Represents a proxy deployed as a sidecar.
            GRPC_SERVER (2):
                Represents a proxyless gRPC backend.
        """
        ENDPOINT_POLICY_TYPE_UNSPECIFIED = 0
        SIDECAR_PROXY = 1
        GRPC_SERVER = 2

    name: str = proto.Field(
        proto.STRING,
        number=1,
    )
    create_time: timestamp_pb2.Timestamp = proto.Field(
        proto.MESSAGE,
        number=2,
        message=timestamp_pb2.Timestamp,
    )
    update_time: timestamp_pb2.Timestamp = proto.Field(
        proto.MESSAGE,
        number=3,
        message=timestamp_pb2.Timestamp,
    )
    labels: MutableMapping[str, str] = proto.MapField(
        proto.STRING,
        proto.STRING,
        number=4,
    )
    type_: EndpointPolicyType = proto.Field(
        proto.ENUM,
        number=5,
        enum=EndpointPolicyType,
    )
    authorization_policy: str = proto.Field(
        proto.STRING,
        number=7,
    )
    endpoint_matcher: 'EndpointMatcher' = proto.Field(
        proto.MESSAGE,
        number=9,
        message='EndpointMatcher',
    )
    traffic_port_selector: 'TrafficPortSelector' = proto.Field(
        proto.MESSAGE,
        number=10,
        message='TrafficPortSelector',
    )
    description: str = proto.Field(
        proto.STRING,
        number=11,
    )
    server_tls_policy: str = proto.Field(
        proto.STRING,
        number=12,
    )
    client_tls_policy: str = proto.Field(
        proto.STRING,
        number=13,
    )


class Gateway(proto.Message):
    r"""Gateway represents the configuration for a proxy, typically a
    load balancer. It captures the ip:port over which the services
    are exposed by the proxy, along with any policy configurations.
    Routes have reference to to Gateways to dictate how requests
    should be routed by this Gateway.

    Attributes:
        name (str):
            Required. Name of the Gateway resource. It matches pattern
            ``projects/*/locations/*/gateways/<gateway_name>``.
        self_link (str):
            Output only. Server-defined URL of this
            resource
        create_time (google.protobuf.timestamp_pb2.Timestamp):
            Output only. The timestamp when the resource
            was created.
        update_time (google.protobuf.timestamp_pb2.Timestamp):
            Output only. The timestamp when the resource
            was updated.
        labels (MutableMapping[str, str]):
            Optional. Set of label tags associated with
            the Gateway resource.
        description (str):
            Optional. A free-text description of the
            resource. Max length 1024 characters.
        type_ (google.events.cloud.networkservices_v1.types.Gateway.Type):
            Immutable. The type of the customer managed
            gateway. This field is required. If unspecified,
            an error is returned.
        addresses (MutableSequence[str]):
            Optional. Zero or one IPv4 or IPv6 address on which the
            Gateway will receive the traffic. When no address is
            provided, an IP from the subnetwork is allocated

            This field only applies to gateways of type
            'SECURE_WEB_GATEWAY'. Gateways of type 'OPEN_MESH' listen on
            0.0.0.0 for IPv4 and :: for IPv6.
        ports (MutableSequence[int]):
            Required. One or more port numbers (1-65535), on which the
            Gateway will receive traffic. The proxy binds to the
            specified ports. Gateways of type 'SECURE_WEB_GATEWAY' are
            limited to 1 port. Gateways of type 'OPEN_MESH' listen on
            0.0.0.0 for IPv4 and :: for IPv6 and support multiple ports.
        scope (str):
            Optional. Scope determines how configuration
            across multiple Gateway instances are merged.
            The configuration for multiple Gateway instances
            with the same scope will be merged as presented
            as a single coniguration to the proxy/load
            balancer.

            Max length 64 characters.
            Scope should start with a letter and can only
            have letters, numbers, hyphens.
        server_tls_policy (str):
            Optional. A fully-qualified ServerTLSPolicy
            URL reference. Specifies how TLS traffic is
            terminated. If empty, TLS termination is
            disabled.
        certificate_urls (MutableSequence[str]):
            Optional. A fully-qualified Certificates URL reference. The
            proxy presents a Certificate (selected based on SNI) when
            establishing a TLS connection. This feature only applies to
            gateways of type 'SECURE_WEB_GATEWAY'.
        gateway_security_policy (str):
            Optional. A fully-qualified GatewaySecurityPolicy URL
            reference. Defines how a server should apply security policy
            to inbound (VM to Proxy) initiated connections.

            For example:
            ``projects/*/locations/*/gatewaySecurityPolicies/swg-policy``.

            This policy is specific to gateways of type
            'SECURE_WEB_GATEWAY'.
        network (str):
            Optional. The relative resource name identifying the VPC
            network that is using this configuration. For example:
            ``projects/*/global/networks/network-1``.

            Currently, this field is specific to gateways of type
            'SECURE_WEB_GATEWAY'.
        subnetwork (str):
            Optional. The relative resource name identifying the
            subnetwork in which this SWG is allocated. For example:
            ``projects/*/regions/us-central1/subnetworks/network-1``

            Currently, this field is specific to gateways of type
            'SECURE_WEB_GATEWAY".
    """
    class Type(proto.Enum):
        r"""The type of the customer-managed gateway. Possible values are:

        -  OPEN_MESH
        -  SECURE_WEB_GATEWAY

        Values:
            TYPE_UNSPECIFIED (0):
                The type of the customer managed gateway is
                unspecified.
            OPEN_MESH (1):
                The type of the customer managed gateway is
                TrafficDirector Open Mesh.
            SECURE_WEB_GATEWAY (2):
                The type of the customer managed gateway is
                SecureWebGateway (SWG).
        """
        TYPE_UNSPECIFIED = 0
        OPEN_MESH = 1
        SECURE_WEB_GATEWAY = 2

    name: str = proto.Field(
        proto.STRING,
        number=1,
    )
    self_link: str = proto.Field(
        proto.STRING,
        number=13,
    )
    create_time: timestamp_pb2.Timestamp = proto.Field(
        proto.MESSAGE,
        number=2,
        message=timestamp_pb2.Timestamp,
    )
    update_time: timestamp_pb2.Timestamp = proto.Field(
        proto.MESSAGE,
        number=3,
        message=timestamp_pb2.Timestamp,
    )
    labels: MutableMapping[str, str] = proto.MapField(
        proto.STRING,
        proto.STRING,
        number=4,
    )
    description: str = proto.Field(
        proto.STRING,
        number=5,
    )
    type_: Type = proto.Field(
        proto.ENUM,
        number=6,
        enum=Type,
    )
    addresses: MutableSequence[str] = proto.RepeatedField(
        proto.STRING,
        number=7,
    )
    ports: MutableSequence[int] = proto.RepeatedField(
        proto.INT32,
        number=11,
    )
    scope: str = proto.Field(
        proto.STRING,
        number=8,
    )
    server_tls_policy: str = proto.Field(
        proto.STRING,
        number=9,
    )
    certificate_urls: MutableSequence[str] = proto.RepeatedField(
        proto.STRING,
        number=14,
    )
    gateway_security_policy: str = proto.Field(
        proto.STRING,
        number=18,
    )
    network: str = proto.Field(
        proto.STRING,
        number=16,
    )
    subnetwork: str = proto.Field(
        proto.STRING,
        number=17,
    )


class GrpcRoute(proto.Message):
    r"""GrpcRoute is the resource defining how gRPC traffic routed by
    a Mesh or Gateway resource is routed.

    Attributes:
        name (str):
            Required. Name of the GrpcRoute resource. It matches pattern
            ``projects/*/locations/global/grpcRoutes/<grpc_route_name>``
        self_link (str):
            Output only. Server-defined URL of this
            resource
        create_time (google.protobuf.timestamp_pb2.Timestamp):
            Output only. The timestamp when the resource
            was created.
        update_time (google.protobuf.timestamp_pb2.Timestamp):
            Output only. The timestamp when the resource
            was updated.
        labels (MutableMapping[str, str]):
            Optional. Set of label tags associated with
            the GrpcRoute resource.
        description (str):
            Optional. A free-text description of the
            resource. Max length 1024 characters.
        hostnames (MutableSequence[str]):
            Required. Service hostnames with an optional port for which
            this route describes traffic.

            Format: [:]

            Hostname is the fully qualified domain name of a network
            host. This matches the RFC 1123 definition of a hostname
            with 2 notable exceptions:

            -  IPs are not allowed.
            -  A hostname may be prefixed with a wildcard label
               (``*.``). The wildcard label must appear by itself as the
               first label.

            Hostname can be "precise" which is a domain name without the
            terminating dot of a network host (e.g. ``foo.example.com``)
            or "wildcard", which is a domain name prefixed with a single
            wildcard label (e.g. ``*.example.com``).

            Note that as per RFC1035 and RFC1123, a label must consist
            of lower case alphanumeric characters or '-', and must start
            and end with an alphanumeric character. No other punctuation
            is allowed.

            The routes associated with a Mesh or Gateway must have
            unique hostnames. If you attempt to attach multiple routes
            with conflicting hostnames, the configuration will be
            rejected.

            For example, while it is acceptable for routes for the
            hostnames ``*.foo.bar.com`` and ``*.bar.com`` to be
            associated with the same route, it is not possible to
            associate two routes both with ``*.bar.com`` or both with
            ``bar.com``.

            If a port is specified, then gRPC clients must use the
            channel URI with the port to match this rule (i.e.
            "xds:///service:123"), otherwise they must supply the URI
            without a port (i.e. "xds:///service").
        meshes (MutableSequence[str]):
            Optional. Meshes defines a list of meshes this GrpcRoute is
            attached to, as one of the routing rules to route the
            requests served by the mesh.

            Each mesh reference should match the pattern:
            ``projects/*/locations/global/meshes/<mesh_name>``
        gateways (MutableSequence[str]):
            Optional. Gateways defines a list of gateways this GrpcRoute
            is attached to, as one of the routing rules to route the
            requests served by the gateway.

            Each gateway reference should match the pattern:
            ``projects/*/locations/global/gateways/<gateway_name>``
        rules (MutableSequence[google.events.cloud.networkservices_v1.types.GrpcRoute.RouteRule]):
            Required. A list of detailed rules defining
            how to route traffic.
            Within a single GrpcRoute, the
            GrpcRoute.RouteAction associated with the first
            matching GrpcRoute.RouteRule will be executed.
            At least one rule must be supplied.
    """

    class MethodMatch(proto.Message):
        r"""Specifies a match against a method.

        .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

        Attributes:
            type_ (google.events.cloud.networkservices_v1.types.GrpcRoute.MethodMatch.Type):
                Optional. Specifies how to match against the
                name. If not specified, a default value of
                "EXACT" is used.
            grpc_service (str):
                Required. Name of the service to match
                against. If unspecified, will match all
                services.
            grpc_method (str):
                Required. Name of the method to match
                against. If unspecified, will match all methods.
            case_sensitive (bool):
                Optional. Specifies that matches are case sensitive. The
                default value is true. case_sensitive must not be used with
                a type of REGULAR_EXPRESSION.

                This field is a member of `oneof`_ ``_case_sensitive``.
        """
        class Type(proto.Enum):
            r"""The type of the match.

            Values:
                TYPE_UNSPECIFIED (0):
                    Unspecified.
                EXACT (1):
                    Will only match the exact name provided.
                REGULAR_EXPRESSION (2):
                    Will interpret grpc_method and grpc_service as regexes. RE2
                    syntax is supported.
            """
            TYPE_UNSPECIFIED = 0
            EXACT = 1
            REGULAR_EXPRESSION = 2

        type_: 'GrpcRoute.MethodMatch.Type' = proto.Field(
            proto.ENUM,
            number=1,
            enum='GrpcRoute.MethodMatch.Type',
        )
        grpc_service: str = proto.Field(
            proto.STRING,
            number=2,
        )
        grpc_method: str = proto.Field(
            proto.STRING,
            number=3,
        )
        case_sensitive: bool = proto.Field(
            proto.BOOL,
            number=4,
            optional=True,
        )

    class HeaderMatch(proto.Message):
        r"""A match against a collection of headers.

        Attributes:
            type_ (google.events.cloud.networkservices_v1.types.GrpcRoute.HeaderMatch.Type):
                Optional. Specifies how to match against the
                value of the header. If not specified, a default
                value of EXACT is used.
            key (str):
                Required. The key of the header.
            value (str):
                Required. The value of the header.
        """
        class Type(proto.Enum):
            r"""The type of match.

            Values:
                TYPE_UNSPECIFIED (0):
                    Unspecified.
                EXACT (1):
                    Will only match the exact value provided.
                REGULAR_EXPRESSION (2):
                    Will match paths conforming to the prefix
                    specified by value. RE2 syntax is supported.
            """
            TYPE_UNSPECIFIED = 0
            EXACT = 1
            REGULAR_EXPRESSION = 2

        type_: 'GrpcRoute.HeaderMatch.Type' = proto.Field(
            proto.ENUM,
            number=1,
            enum='GrpcRoute.HeaderMatch.Type',
        )
        key: str = proto.Field(
            proto.STRING,
            number=2,
        )
        value: str = proto.Field(
            proto.STRING,
            number=3,
        )

    class RouteMatch(proto.Message):
        r"""Criteria for matching traffic. A RouteMatch will be
        considered to match when all supplied fields match.


        .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

        Attributes:
            method (google.events.cloud.networkservices_v1.types.GrpcRoute.MethodMatch):
                Optional. A gRPC method to match against. If
                this field is empty or omitted, will match all
                methods.

                This field is a member of `oneof`_ ``_method``.
            headers (MutableSequence[google.events.cloud.networkservices_v1.types.GrpcRoute.HeaderMatch]):
                Optional. Specifies a collection of headers
                to match.
        """

        method: 'GrpcRoute.MethodMatch' = proto.Field(
            proto.MESSAGE,
            number=1,
            optional=True,
            message='GrpcRoute.MethodMatch',
        )
        headers: MutableSequence['GrpcRoute.HeaderMatch'] = proto.RepeatedField(
            proto.MESSAGE,
            number=2,
            message='GrpcRoute.HeaderMatch',
        )

    class Destination(proto.Message):
        r"""The destination to which traffic will be routed.

        .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

        Attributes:
            service_name (str):
                Required. The URL of a destination service to
                which to route traffic. Must refer to either a
                BackendService or ServiceDirectoryService.

                This field is a member of `oneof`_ ``destination_type``.
            weight (int):
                Optional. Specifies the proportion of
                requests forwarded to the backend referenced by
                the serviceName field. This is computed as:

                        weight/Sum(weights in this destination
                list). For non-zero values, there may be some
                epsilon from the exact proportion defined here
                depending on the precision an implementation
                supports.

                If only one serviceName is specified and it has
                a weight greater than 0, 100% of the traffic is
                forwarded to that backend.

                If weights are specified for any one service
                name, they need to be specified for all of them.

                If weights are unspecified for all services,
                then, traffic is distributed in equal
                proportions to all of them.

                This field is a member of `oneof`_ ``_weight``.
        """

        service_name: str = proto.Field(
            proto.STRING,
            number=1,
            oneof='destination_type',
        )
        weight: int = proto.Field(
            proto.INT32,
            number=2,
            optional=True,
        )

    class FaultInjectionPolicy(proto.Message):
        r"""The specification for fault injection introduced into traffic
        to test the resiliency of clients to destination service
        failure. As part of fault injection, when clients send requests
        to a destination, delays can be introduced on a percentage of
        requests before sending those requests to the destination
        service. Similarly requests from clients can be aborted by for a
        percentage of requests.


        .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

        Attributes:
            delay (google.events.cloud.networkservices_v1.types.GrpcRoute.FaultInjectionPolicy.Delay):
                The specification for injecting delay to
                client requests.

                This field is a member of `oneof`_ ``_delay``.
            abort (google.events.cloud.networkservices_v1.types.GrpcRoute.FaultInjectionPolicy.Abort):
                The specification for aborting to client
                requests.

                This field is a member of `oneof`_ ``_abort``.
        """

        class Delay(proto.Message):
            r"""Specification of how client requests are delayed as part of
            fault injection before being sent to a destination.


            .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

            Attributes:
                fixed_delay (google.protobuf.duration_pb2.Duration):
                    Specify a fixed delay before forwarding the
                    request.

                    This field is a member of `oneof`_ ``_fixed_delay``.
                percentage (int):
                    The percentage of traffic on which delay will be injected.

                    The value must be between [0, 100]

                    This field is a member of `oneof`_ ``_percentage``.
            """

            fixed_delay: duration_pb2.Duration = proto.Field(
                proto.MESSAGE,
                number=1,
                optional=True,
                message=duration_pb2.Duration,
            )
            percentage: int = proto.Field(
                proto.INT32,
                number=2,
                optional=True,
            )

        class Abort(proto.Message):
            r"""Specification of how client requests are aborted as part of
            fault injection before being sent to a destination.


            .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

            Attributes:
                http_status (int):
                    The HTTP status code used to abort the
                    request.
                    The value must be between 200 and 599 inclusive.

                    This field is a member of `oneof`_ ``_http_status``.
                percentage (int):
                    The percentage of traffic which will be aborted.

                    The value must be between [0, 100]

                    This field is a member of `oneof`_ ``_percentage``.
            """

            http_status: int = proto.Field(
                proto.INT32,
                number=1,
                optional=True,
            )
            percentage: int = proto.Field(
                proto.INT32,
                number=2,
                optional=True,
            )

        delay: 'GrpcRoute.FaultInjectionPolicy.Delay' = proto.Field(
            proto.MESSAGE,
            number=1,
            optional=True,
            message='GrpcRoute.FaultInjectionPolicy.Delay',
        )
        abort: 'GrpcRoute.FaultInjectionPolicy.Abort' = proto.Field(
            proto.MESSAGE,
            number=2,
            optional=True,
            message='GrpcRoute.FaultInjectionPolicy.Abort',
        )

    class RetryPolicy(proto.Message):
        r"""The specifications for retries.

        Attributes:
            retry_conditions (MutableSequence[str]):
                -  connect-failure: Router will retry on failures connecting
                   to Backend Services, for example due to connection
                   timeouts.
                -  refused-stream: Router will retry if the backend service
                   resets the stream with a REFUSED_STREAM error code. This
                   reset type indicates that it is safe to retry.
                -  cancelled: Router will retry if the gRPC status code in
                   the response header is set to cancelled
                -  deadline-exceeded: Router will retry if the gRPC status
                   code in the response header is set to deadline-exceeded
                -  resource-exhausted: Router will retry if the gRPC status
                   code in the response header is set to resource-exhausted
                -  unavailable: Router will retry if the gRPC status code in
                   the response header is set to unavailable
            num_retries (int):
                Specifies the allowed number of retries. This
                number must be > 0. If not specified, default to
                1.
        """

        retry_conditions: MutableSequence[str] = proto.RepeatedField(
            proto.STRING,
            number=1,
        )
        num_retries: int = proto.Field(
            proto.UINT32,
            number=2,
        )

    class RouteAction(proto.Message):
        r"""Specifies how to route matched traffic.

        Attributes:
            destinations (MutableSequence[google.events.cloud.networkservices_v1.types.GrpcRoute.Destination]):
                Optional. The destination services to which
                traffic should be forwarded. If multiple
                destinations are specified, traffic will be
                split between Backend Service(s) according to
                the weight field of these destinations.
            fault_injection_policy (google.events.cloud.networkservices_v1.types.GrpcRoute.FaultInjectionPolicy):
                Optional. The specification for fault injection introduced
                into traffic to test the resiliency of clients to
                destination service failure. As part of fault injection,
                when clients send requests to a destination, delays can be
                introduced on a percentage of requests before sending those
                requests to the destination service. Similarly requests from
                clients can be aborted by for a percentage of requests.

                timeout and retry_policy will be ignored by clients that are
                configured with a fault_injection_policy
            timeout (google.protobuf.duration_pb2.Duration):
                Optional. Specifies the timeout for selected
                route. Timeout is computed from the time the
                request has been fully processed (i.e. end of
                stream) up until the response has been
                completely processed. Timeout includes all
                retries.
            retry_policy (google.events.cloud.networkservices_v1.types.GrpcRoute.RetryPolicy):
                Optional. Specifies the retry policy
                associated with this route.
        """

        destinations: MutableSequence['GrpcRoute.Destination'] = proto.RepeatedField(
            proto.MESSAGE,
            number=1,
            message='GrpcRoute.Destination',
        )
        fault_injection_policy: 'GrpcRoute.FaultInjectionPolicy' = proto.Field(
            proto.MESSAGE,
            number=3,
            message='GrpcRoute.FaultInjectionPolicy',
        )
        timeout: duration_pb2.Duration = proto.Field(
            proto.MESSAGE,
            number=7,
            message=duration_pb2.Duration,
        )
        retry_policy: 'GrpcRoute.RetryPolicy' = proto.Field(
            proto.MESSAGE,
            number=8,
            message='GrpcRoute.RetryPolicy',
        )

    class RouteRule(proto.Message):
        r"""Describes how to route traffic.

        Attributes:
            matches (MutableSequence[google.events.cloud.networkservices_v1.types.GrpcRoute.RouteMatch]):
                Optional. Matches define conditions used for
                matching the rule against incoming gRPC
                requests. Each match is independent, i.e. this
                rule will be matched if ANY one of the matches
                is satisfied.  If no matches field is specified,
                this rule will unconditionally match traffic.
            action (google.events.cloud.networkservices_v1.types.GrpcRoute.RouteAction):
                Required. A detailed rule defining how to
                route traffic. This field is required.
        """

        matches: MutableSequence['GrpcRoute.RouteMatch'] = proto.RepeatedField(
            proto.MESSAGE,
            number=1,
            message='GrpcRoute.RouteMatch',
        )
        action: 'GrpcRoute.RouteAction' = proto.Field(
            proto.MESSAGE,
            number=2,
            message='GrpcRoute.RouteAction',
        )

    name: str = proto.Field(
        proto.STRING,
        number=1,
    )
    self_link: str = proto.Field(
        proto.STRING,
        number=12,
    )
    create_time: timestamp_pb2.Timestamp = proto.Field(
        proto.MESSAGE,
        number=2,
        message=timestamp_pb2.Timestamp,
    )
    update_time: timestamp_pb2.Timestamp = proto.Field(
        proto.MESSAGE,
        number=3,
        message=timestamp_pb2.Timestamp,
    )
    labels: MutableMapping[str, str] = proto.MapField(
        proto.STRING,
        proto.STRING,
        number=4,
    )
    description: str = proto.Field(
        proto.STRING,
        number=5,
    )
    hostnames: MutableSequence[str] = proto.RepeatedField(
        proto.STRING,
        number=6,
    )
    meshes: MutableSequence[str] = proto.RepeatedField(
        proto.STRING,
        number=9,
    )
    gateways: MutableSequence[str] = proto.RepeatedField(
        proto.STRING,
        number=10,
    )
    rules: MutableSequence[RouteRule] = proto.RepeatedField(
        proto.MESSAGE,
        number=7,
        message=RouteRule,
    )


class HttpRoute(proto.Message):
    r"""HttpRoute is the resource defining how HTTP traffic should be
    routed by a Mesh or Gateway resource.

    Attributes:
        name (str):
            Required. Name of the HttpRoute resource. It matches pattern
            ``projects/*/locations/global/httpRoutes/http_route_name>``.
        self_link (str):
            Output only. Server-defined URL of this
            resource
        description (str):
            Optional. A free-text description of the
            resource. Max length 1024 characters.
        create_time (google.protobuf.timestamp_pb2.Timestamp):
            Output only. The timestamp when the resource
            was created.
        update_time (google.protobuf.timestamp_pb2.Timestamp):
            Output only. The timestamp when the resource
            was updated.
        hostnames (MutableSequence[str]):
            Required. Hostnames define a set of hosts that should match
            against the HTTP host header to select a HttpRoute to
            process the request. Hostname is the fully qualified domain
            name of a network host, as defined by RFC 1123 with the
            exception that:

            -  IPs are not allowed.
            -  A hostname may be prefixed with a wildcard label
               (``*.``). The wildcard label must appear by itself as the
               first label.

            Hostname can be "precise" which is a domain name without the
            terminating dot of a network host (e.g. ``foo.example.com``)
            or "wildcard", which is a domain name prefixed with a single
            wildcard label (e.g. ``*.example.com``).

            Note that as per RFC1035 and RFC1123, a label must consist
            of lower case alphanumeric characters or '-', and must start
            and end with an alphanumeric character. No other punctuation
            is allowed.

            The routes associated with a Mesh or Gateways must have
            unique hostnames. If you attempt to attach multiple routes
            with conflicting hostnames, the configuration will be
            rejected.

            For example, while it is acceptable for routes for the
            hostnames ``*.foo.bar.com`` and ``*.bar.com`` to be
            associated with the same Mesh (or Gateways under the same
            scope), it is not possible to associate two routes both with
            ``*.bar.com`` or both with ``bar.com``.
        meshes (MutableSequence[str]):
            Optional. Meshes defines a list of meshes this HttpRoute is
            attached to, as one of the routing rules to route the
            requests served by the mesh.

            Each mesh reference should match the pattern:
            ``projects/*/locations/global/meshes/<mesh_name>``

            The attached Mesh should be of a type SIDECAR
        gateways (MutableSequence[str]):
            Optional. Gateways defines a list of gateways this HttpRoute
            is attached to, as one of the routing rules to route the
            requests served by the gateway.

            Each gateway reference should match the pattern:
            ``projects/*/locations/global/gateways/<gateway_name>``
        labels (MutableMapping[str, str]):
            Optional. Set of label tags associated with
            the HttpRoute resource.
        rules (MutableSequence[google.events.cloud.networkservices_v1.types.HttpRoute.RouteRule]):
            Required. Rules that define how traffic is
            routed and handled. Rules will be matched
            sequentially based on the RouteMatch specified
            for the rule.
    """

    class HeaderMatch(proto.Message):
        r"""Specifies how to select a route rule based on HTTP request
        headers.

        This message has `oneof`_ fields (mutually exclusive fields).
        For each oneof, at most one member field can be set at the same time.
        Setting any member of the oneof automatically clears all other
        members.

        .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

        Attributes:
            exact_match (str):
                The value of the header should match exactly the content of
                exact_match.

                This field is a member of `oneof`_ ``MatchType``.
            regex_match (str):
                The value of the header must match the regular expression
                specified in regex_match. For regular expression grammar,
                please see: https://github.com/google/re2/wiki/Syntax

                This field is a member of `oneof`_ ``MatchType``.
            prefix_match (str):
                The value of the header must start with the contents of
                prefix_match.

                This field is a member of `oneof`_ ``MatchType``.
            present_match (bool):
                A header with header_name must exist. The match takes place
                whether or not the header has a value.

                This field is a member of `oneof`_ ``MatchType``.
            suffix_match (str):
                The value of the header must end with the contents of
                suffix_match.

                This field is a member of `oneof`_ ``MatchType``.
            range_match (google.events.cloud.networkservices_v1.types.HttpRoute.HeaderMatch.IntegerRange):
                If specified, the rule will match if the
                request header value is within the range.

                This field is a member of `oneof`_ ``MatchType``.
            header (str):
                The name of the HTTP header to match against.
            invert_match (bool):
                If specified, the match result will be
                inverted before checking. Default value is set
                to false.
        """

        class IntegerRange(proto.Message):
            r"""Represents an integer value range.

            Attributes:
                start (int):
                    Start of the range (inclusive)
                end (int):
                    End of the range (exclusive)
            """

            start: int = proto.Field(
                proto.INT32,
                number=1,
            )
            end: int = proto.Field(
                proto.INT32,
                number=2,
            )

        exact_match: str = proto.Field(
            proto.STRING,
            number=2,
            oneof='MatchType',
        )
        regex_match: str = proto.Field(
            proto.STRING,
            number=3,
            oneof='MatchType',
        )
        prefix_match: str = proto.Field(
            proto.STRING,
            number=4,
            oneof='MatchType',
        )
        present_match: bool = proto.Field(
            proto.BOOL,
            number=5,
            oneof='MatchType',
        )
        suffix_match: str = proto.Field(
            proto.STRING,
            number=6,
            oneof='MatchType',
        )
        range_match: 'HttpRoute.HeaderMatch.IntegerRange' = proto.Field(
            proto.MESSAGE,
            number=7,
            oneof='MatchType',
            message='HttpRoute.HeaderMatch.IntegerRange',
        )
        header: str = proto.Field(
            proto.STRING,
            number=1,
        )
        invert_match: bool = proto.Field(
            proto.BOOL,
            number=8,
        )

    class QueryParameterMatch(proto.Message):
        r"""Specifications to match a query parameter in the request.

        This message has `oneof`_ fields (mutually exclusive fields).
        For each oneof, at most one member field can be set at the same time.
        Setting any member of the oneof automatically clears all other
        members.

        .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

        Attributes:
            exact_match (str):
                The value of the query parameter must exactly match the
                contents of exact_match.

                Only one of exact_match, regex_match, or present_match must
                be set.

                This field is a member of `oneof`_ ``MatchType``.
            regex_match (str):
                The value of the query parameter must match the regular
                expression specified by regex_match. For regular expression
                grammar, please see
                https://github.com/google/re2/wiki/Syntax

                Only one of exact_match, regex_match, or present_match must
                be set.

                This field is a member of `oneof`_ ``MatchType``.
            present_match (bool):
                Specifies that the QueryParameterMatcher matches if request
                contains query parameter, irrespective of whether the
                parameter has a value or not.

                Only one of exact_match, regex_match, or present_match must
                be set.

                This field is a member of `oneof`_ ``MatchType``.
            query_parameter (str):
                The name of the query parameter to match.
        """

        exact_match: str = proto.Field(
            proto.STRING,
            number=2,
            oneof='MatchType',
        )
        regex_match: str = proto.Field(
            proto.STRING,
            number=3,
            oneof='MatchType',
        )
        present_match: bool = proto.Field(
            proto.BOOL,
            number=4,
            oneof='MatchType',
        )
        query_parameter: str = proto.Field(
            proto.STRING,
            number=1,
        )

    class RouteMatch(proto.Message):
        r"""RouteMatch defines specifications used to match requests. If
        multiple match types are set, this RouteMatch will match if ALL
        type of matches are matched.

        This message has `oneof`_ fields (mutually exclusive fields).
        For each oneof, at most one member field can be set at the same time.
        Setting any member of the oneof automatically clears all other
        members.

        .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

        Attributes:
            full_path_match (str):
                The HTTP request path value should exactly match this value.

                Only one of full_path_match, prefix_match, or regex_match
                should be used.

                This field is a member of `oneof`_ ``PathMatch``.
            prefix_match (str):
                The HTTP request path value must begin with specified
                prefix_match. prefix_match must begin with a /.

                Only one of full_path_match, prefix_match, or regex_match
                should be used.

                This field is a member of `oneof`_ ``PathMatch``.
            regex_match (str):
                The HTTP request path value must satisfy the regular
                expression specified by regex_match after removing any query
                parameters and anchor supplied with the original URL. For
                regular expression grammar, please see
                https://github.com/google/re2/wiki/Syntax

                Only one of full_path_match, prefix_match, or regex_match
                should be used.

                This field is a member of `oneof`_ ``PathMatch``.
            ignore_case (bool):
                Specifies if prefix_match and full_path_match matches are
                case sensitive. The default value is false.
            headers (MutableSequence[google.events.cloud.networkservices_v1.types.HttpRoute.HeaderMatch]):
                Specifies a list of HTTP request headers to
                match against. ALL of the supplied headers must
                be matched.
            query_parameters (MutableSequence[google.events.cloud.networkservices_v1.types.HttpRoute.QueryParameterMatch]):
                Specifies a list of query parameters to match
                against. ALL of the query parameters must be
                matched.
        """

        full_path_match: str = proto.Field(
            proto.STRING,
            number=1,
            oneof='PathMatch',
        )
        prefix_match: str = proto.Field(
            proto.STRING,
            number=2,
            oneof='PathMatch',
        )
        regex_match: str = proto.Field(
            proto.STRING,
            number=3,
            oneof='PathMatch',
        )
        ignore_case: bool = proto.Field(
            proto.BOOL,
            number=4,
        )
        headers: MutableSequence['HttpRoute.HeaderMatch'] = proto.RepeatedField(
            proto.MESSAGE,
            number=5,
            message='HttpRoute.HeaderMatch',
        )
        query_parameters: MutableSequence['HttpRoute.QueryParameterMatch'] = proto.RepeatedField(
            proto.MESSAGE,
            number=6,
            message='HttpRoute.QueryParameterMatch',
        )

    class Destination(proto.Message):
        r"""Specifications of a destination to which the request should
        be routed to.

        Attributes:
            service_name (str):
                The URL of a BackendService to route traffic
                to.
            weight (int):
                Specifies the proportion of requests
                forwarded to the backend referenced by the
                serviceName field. This is computed as:

                        weight/Sum(weights in this destination
                list). For non-zero values, there may be some
                epsilon from the exact proportion defined here
                depending on the precision an implementation
                supports.

                If only one serviceName is specified and it has
                a weight greater than 0, 100% of the traffic is
                forwarded to that backend.

                If weights are specified for any one service
                name, they need to be specified for all of them.

                If weights are unspecified for all services,
                then, traffic is distributed in equal
                proportions to all of them.
        """

        service_name: str = proto.Field(
            proto.STRING,
            number=1,
        )
        weight: int = proto.Field(
            proto.INT32,
            number=2,
        )

    class Redirect(proto.Message):
        r"""The specification for redirecting traffic.

        Attributes:
            host_redirect (str):
                The host that will be used in the redirect
                response instead of the one that was supplied in
                the request.
            path_redirect (str):
                The path that will be used in the redirect response instead
                of the one that was supplied in the request. path_redirect
                can not be supplied together with prefix_redirect. Supply
                one alone or neither. If neither is supplied, the path of
                the original request will be used for the redirect.
            prefix_rewrite (str):
                Indicates that during redirection, the
                matched prefix (or path) should be swapped with
                this value. This option allows URLs be
                dynamically created based on the request.
            response_code (google.events.cloud.networkservices_v1.types.HttpRoute.Redirect.ResponseCode):
                The HTTP Status code to use for the redirect.
            https_redirect (bool):
                If set to true, the URL scheme in the
                redirected request is set to https. If set to
                false, the URL scheme of the redirected request
                will remain the same as that of the request.

                The default is set to false.
            strip_query (bool):
                if set to true, any accompanying query
                portion of the original URL is removed prior to
                redirecting the request. If set to false, the
                query portion of the original URL is retained.

                The default is set to false.
            port_redirect (int):
                The port that will be used in the redirected
                request instead of the one that was supplied in
                the request.
        """
        class ResponseCode(proto.Enum):
            r"""Supported HTTP response code.

            Values:
                RESPONSE_CODE_UNSPECIFIED (0):
                    Default value
                MOVED_PERMANENTLY_DEFAULT (1):
                    Corresponds to 301.
                FOUND (2):
                    Corresponds to 302.
                SEE_OTHER (3):
                    Corresponds to 303.
                TEMPORARY_REDIRECT (4):
                    Corresponds to 307. In this case, the request
                    method will be retained.
                PERMANENT_REDIRECT (5):
                    Corresponds to 308. In this case, the request
                    method will be retained.
            """
            RESPONSE_CODE_UNSPECIFIED = 0
            MOVED_PERMANENTLY_DEFAULT = 1
            FOUND = 2
            SEE_OTHER = 3
            TEMPORARY_REDIRECT = 4
            PERMANENT_REDIRECT = 5

        host_redirect: str = proto.Field(
            proto.STRING,
            number=1,
        )
        path_redirect: str = proto.Field(
            proto.STRING,
            number=2,
        )
        prefix_rewrite: str = proto.Field(
            proto.STRING,
            number=3,
        )
        response_code: 'HttpRoute.Redirect.ResponseCode' = proto.Field(
            proto.ENUM,
            number=4,
            enum='HttpRoute.Redirect.ResponseCode',
        )
        https_redirect: bool = proto.Field(
            proto.BOOL,
            number=5,
        )
        strip_query: bool = proto.Field(
            proto.BOOL,
            number=6,
        )
        port_redirect: int = proto.Field(
            proto.INT32,
            number=7,
        )

    class FaultInjectionPolicy(proto.Message):
        r"""The specification for fault injection introduced into traffic
        to test the resiliency of clients to destination service
        failure. As part of fault injection, when clients send requests
        to a destination, delays can be introduced by client proxy on a
        percentage of requests before sending those requests to the
        destination service. Similarly requests can be aborted by client
        proxy for a percentage of requests.

        Attributes:
            delay (google.events.cloud.networkservices_v1.types.HttpRoute.FaultInjectionPolicy.Delay):
                The specification for injecting delay to
                client requests.
            abort (google.events.cloud.networkservices_v1.types.HttpRoute.FaultInjectionPolicy.Abort):
                The specification for aborting to client
                requests.
        """

        class Delay(proto.Message):
            r"""Specification of how client requests are delayed as part of
            fault injection before being sent to a destination.

            Attributes:
                fixed_delay (google.protobuf.duration_pb2.Duration):
                    Specify a fixed delay before forwarding the
                    request.
                percentage (int):
                    The percentage of traffic on which delay will be injected.

                    The value must be between [0, 100]
            """

            fixed_delay: duration_pb2.Duration = proto.Field(
                proto.MESSAGE,
                number=1,
                message=duration_pb2.Duration,
            )
            percentage: int = proto.Field(
                proto.INT32,
                number=2,
            )

        class Abort(proto.Message):
            r"""Specification of how client requests are aborted as part of
            fault injection before being sent to a destination.

            Attributes:
                http_status (int):
                    The HTTP status code used to abort the
                    request.
                    The value must be between 200 and 599 inclusive.
                percentage (int):
                    The percentage of traffic which will be aborted.

                    The value must be between [0, 100]
            """

            http_status: int = proto.Field(
                proto.INT32,
                number=1,
            )
            percentage: int = proto.Field(
                proto.INT32,
                number=2,
            )

        delay: 'HttpRoute.FaultInjectionPolicy.Delay' = proto.Field(
            proto.MESSAGE,
            number=1,
            message='HttpRoute.FaultInjectionPolicy.Delay',
        )
        abort: 'HttpRoute.FaultInjectionPolicy.Abort' = proto.Field(
            proto.MESSAGE,
            number=2,
            message='HttpRoute.FaultInjectionPolicy.Abort',
        )

    class HeaderModifier(proto.Message):
        r"""The specification for modifying HTTP header in HTTP request
        and HTTP response.

        Attributes:
            set (MutableMapping[str, str]):
                Completely overwrite/replace the headers with
                given map where key is the name of the header,
                value is the value of the header.
            add (MutableMapping[str, str]):
                Add the headers with given map where key is
                the name of the header, value is the value of
                the header.
            remove (MutableSequence[str]):
                Remove headers (matching by header names)
                specified in the list.
        """

        set: MutableMapping[str, str] = proto.MapField(
            proto.STRING,
            proto.STRING,
            number=1,
        )
        add: MutableMapping[str, str] = proto.MapField(
            proto.STRING,
            proto.STRING,
            number=2,
        )
        remove: MutableSequence[str] = proto.RepeatedField(
            proto.STRING,
            number=3,
        )

    class URLRewrite(proto.Message):
        r"""The specification for modifying the URL of the request, prior
        to forwarding the request to the destination.

        Attributes:
            path_prefix_rewrite (str):
                Prior to forwarding the request to the
                selected destination, the matching portion of
                the requests path is replaced by this value.
            host_rewrite (str):
                Prior to forwarding the request to the
                selected destination, the requests host header
                is replaced by this value.
        """

        path_prefix_rewrite: str = proto.Field(
            proto.STRING,
            number=1,
        )
        host_rewrite: str = proto.Field(
            proto.STRING,
            number=2,
        )

    class RetryPolicy(proto.Message):
        r"""The specifications for retries.

        Attributes:
            retry_conditions (MutableSequence[str]):
                Specifies one or more conditions when this retry policy
                applies. Valid values are: 5xx: Proxy will attempt a retry
                if the destination service responds with any 5xx response
                code, of if the destination service does not respond at all,
                example: disconnect, reset, read timeout, connection failure
                and refused streams.

                gateway-error: Similar to 5xx, but only applies to response
                codes 502, 503, 504.

                reset: Proxy will attempt a retry if the destination service
                does not respond at all (disconnect/reset/read timeout)

                connect-failure: Proxy will retry on failures connecting to
                destination for example due to connection timeouts.

                retriable-4xx: Proxy will retry fro retriable 4xx response
                codes. Currently the only retriable error supported is 409.

                refused-stream: Proxy will retry if the destination resets
                the stream with a REFUSED_STREAM error code. This reset type
                indicates that it is safe to retry.
            num_retries (int):
                Specifies the allowed number of retries. This
                number must be > 0. If not specified, default to
                1.
            per_try_timeout (google.protobuf.duration_pb2.Duration):
                Specifies a non-zero timeout per retry
                attempt.
        """

        retry_conditions: MutableSequence[str] = proto.RepeatedField(
            proto.STRING,
            number=1,
        )
        num_retries: int = proto.Field(
            proto.INT32,
            number=2,
        )
        per_try_timeout: duration_pb2.Duration = proto.Field(
            proto.MESSAGE,
            number=3,
            message=duration_pb2.Duration,
        )

    class RequestMirrorPolicy(proto.Message):
        r"""Specifies the policy on how requests are shadowed to a
        separate mirrored destination service. The proxy does not wait
        for responses from the shadow service. Prior to sending traffic
        to the shadow service, the host/authority header is suffixed
        with -shadow.

        Attributes:
            destination (google.events.cloud.networkservices_v1.types.HttpRoute.Destination):
                The destination the requests will be mirrored
                to. The weight of the destination will be
                ignored.
        """

        destination: 'HttpRoute.Destination' = proto.Field(
            proto.MESSAGE,
            number=1,
            message='HttpRoute.Destination',
        )

    class CorsPolicy(proto.Message):
        r"""The Specification for allowing client side cross-origin
        requests.

        Attributes:
            allow_origins (MutableSequence[str]):
                Specifies the list of origins that will be allowed to do
                CORS requests. An origin is allowed if it matches either an
                item in allow_origins or an item in allow_origin_regexes.
            allow_origin_regexes (MutableSequence[str]):
                Specifies the regular expression patterns
                that match allowed origins. For regular
                expression grammar, please see
                https://github.com/google/re2/wiki/Syntax.
            allow_methods (MutableSequence[str]):
                Specifies the content for
                Access-Control-Allow-Methods header.
            allow_headers (MutableSequence[str]):
                Specifies the content for
                Access-Control-Allow-Headers header.
            expose_headers (MutableSequence[str]):
                Specifies the content for
                Access-Control-Expose-Headers header.
            max_age (str):
                Specifies how long result of a preflight
                request can be cached in seconds. This
                translates to the Access-Control-Max-Age header.
            allow_credentials (bool):
                In response to a preflight request, setting
                this to true indicates that the actual request
                can include user credentials. This translates to
                the Access-Control-Allow-Credentials header.

                Default value is false.
            disabled (bool):
                If true, the CORS policy is disabled. The
                default value is false, which indicates that the
                CORS policy is in effect.
        """

        allow_origins: MutableSequence[str] = proto.RepeatedField(
            proto.STRING,
            number=1,
        )
        allow_origin_regexes: MutableSequence[str] = proto.RepeatedField(
            proto.STRING,
            number=2,
        )
        allow_methods: MutableSequence[str] = proto.RepeatedField(
            proto.STRING,
            number=3,
        )
        allow_headers: MutableSequence[str] = proto.RepeatedField(
            proto.STRING,
            number=4,
        )
        expose_headers: MutableSequence[str] = proto.RepeatedField(
            proto.STRING,
            number=5,
        )
        max_age: str = proto.Field(
            proto.STRING,
            number=6,
        )
        allow_credentials: bool = proto.Field(
            proto.BOOL,
            number=7,
        )
        disabled: bool = proto.Field(
            proto.BOOL,
            number=8,
        )

    class RouteAction(proto.Message):
        r"""The specifications for routing traffic and applying
        associated policies.

        Attributes:
            destinations (MutableSequence[google.events.cloud.networkservices_v1.types.HttpRoute.Destination]):
                The destination to which traffic should be
                forwarded.
            redirect (google.events.cloud.networkservices_v1.types.HttpRoute.Redirect):
                If set, the request is directed as configured
                by this field.
            fault_injection_policy (google.events.cloud.networkservices_v1.types.HttpRoute.FaultInjectionPolicy):
                The specification for fault injection introduced into
                traffic to test the resiliency of clients to backend service
                failure. As part of fault injection, when clients send
                requests to a backend service, delays can be introduced on a
                percentage of requests before sending those requests to the
                backend service. Similarly requests from clients can be
                aborted for a percentage of requests.

                timeout and retry_policy will be ignored by clients that are
                configured with a fault_injection_policy
            request_header_modifier (google.events.cloud.networkservices_v1.types.HttpRoute.HeaderModifier):
                The specification for modifying the headers
                of a matching request prior to delivery of the
                request to the destination.
            response_header_modifier (google.events.cloud.networkservices_v1.types.HttpRoute.HeaderModifier):
                The specification for modifying the headers
                of a response prior to sending the response back
                to the client.
            url_rewrite (google.events.cloud.networkservices_v1.types.HttpRoute.URLRewrite):
                The specification for rewrite URL before
                forwarding requests to the destination.
            timeout (google.protobuf.duration_pb2.Duration):
                Specifies the timeout for selected route.
                Timeout is computed from the time the request
                has been fully processed (i.e. end of stream) up
                until the response has been completely
                processed. Timeout includes all retries.
            retry_policy (google.events.cloud.networkservices_v1.types.HttpRoute.RetryPolicy):
                Specifies the retry policy associated with
                this route.
            request_mirror_policy (google.events.cloud.networkservices_v1.types.HttpRoute.RequestMirrorPolicy):
                Specifies the policy on how requests intended
                for the routes destination are shadowed to a
                separate mirrored destination. Proxy will not
                wait for the shadow destination to respond
                before returning the response. Prior to sending
                traffic to the shadow service, the
                host/authority header is suffixed with -shadow.
            cors_policy (google.events.cloud.networkservices_v1.types.HttpRoute.CorsPolicy):
                The specification for allowing client side
                cross-origin requests.
        """

        destinations: MutableSequence['HttpRoute.Destination'] = proto.RepeatedField(
            proto.MESSAGE,
            number=1,
            message='HttpRoute.Destination',
        )
        redirect: 'HttpRoute.Redirect' = proto.Field(
            proto.MESSAGE,
            number=2,
            message='HttpRoute.Redirect',
        )
        fault_injection_policy: 'HttpRoute.FaultInjectionPolicy' = proto.Field(
            proto.MESSAGE,
            number=4,
            message='HttpRoute.FaultInjectionPolicy',
        )
        request_header_modifier: 'HttpRoute.HeaderModifier' = proto.Field(
            proto.MESSAGE,
            number=5,
            message='HttpRoute.HeaderModifier',
        )
        response_header_modifier: 'HttpRoute.HeaderModifier' = proto.Field(
            proto.MESSAGE,
            number=6,
            message='HttpRoute.HeaderModifier',
        )
        url_rewrite: 'HttpRoute.URLRewrite' = proto.Field(
            proto.MESSAGE,
            number=7,
            message='HttpRoute.URLRewrite',
        )
        timeout: duration_pb2.Duration = proto.Field(
            proto.MESSAGE,
            number=8,
            message=duration_pb2.Duration,
        )
        retry_policy: 'HttpRoute.RetryPolicy' = proto.Field(
            proto.MESSAGE,
            number=9,
            message='HttpRoute.RetryPolicy',
        )
        request_mirror_policy: 'HttpRoute.RequestMirrorPolicy' = proto.Field(
            proto.MESSAGE,
            number=10,
            message='HttpRoute.RequestMirrorPolicy',
        )
        cors_policy: 'HttpRoute.CorsPolicy' = proto.Field(
            proto.MESSAGE,
            number=11,
            message='HttpRoute.CorsPolicy',
        )

    class RouteRule(proto.Message):
        r"""Specifies how to match traffic and how to route traffic when
        traffic is matched.

        Attributes:
            matches (MutableSequence[google.events.cloud.networkservices_v1.types.HttpRoute.RouteMatch]):
                A list of matches define conditions used for
                matching the rule against incoming HTTP
                requests. Each match is independent, i.e. this
                rule will be matched if ANY one of the matches
                is satisfied.

                If no matches field is specified, this rule will
                unconditionally match traffic.

                If a default rule is desired to be configured,
                add a rule with no matches specified to the end
                of the rules list.
            action (google.events.cloud.networkservices_v1.types.HttpRoute.RouteAction):
                The detailed rule defining how to route
                matched traffic.
        """

        matches: MutableSequence['HttpRoute.RouteMatch'] = proto.RepeatedField(
            proto.MESSAGE,
            number=1,
            message='HttpRoute.RouteMatch',
        )
        action: 'HttpRoute.RouteAction' = proto.Field(
            proto.MESSAGE,
            number=2,
            message='HttpRoute.RouteAction',
        )

    name: str = proto.Field(
        proto.STRING,
        number=1,
    )
    self_link: str = proto.Field(
        proto.STRING,
        number=11,
    )
    description: str = proto.Field(
        proto.STRING,
        number=2,
    )
    create_time: timestamp_pb2.Timestamp = proto.Field(
        proto.MESSAGE,
        number=3,
        message=timestamp_pb2.Timestamp,
    )
    update_time: timestamp_pb2.Timestamp = proto.Field(
        proto.MESSAGE,
        number=4,
        message=timestamp_pb2.Timestamp,
    )
    hostnames: MutableSequence[str] = proto.RepeatedField(
        proto.STRING,
        number=5,
    )
    meshes: MutableSequence[str] = proto.RepeatedField(
        proto.STRING,
        number=8,
    )
    gateways: MutableSequence[str] = proto.RepeatedField(
        proto.STRING,
        number=9,
    )
    labels: MutableMapping[str, str] = proto.MapField(
        proto.STRING,
        proto.STRING,
        number=10,
    )
    rules: MutableSequence[RouteRule] = proto.RepeatedField(
        proto.MESSAGE,
        number=6,
        message=RouteRule,
    )


class Mesh(proto.Message):
    r"""Mesh represents a logical configuration grouping for workload
    to workload communication within a service mesh. Routes that
    point to mesh dictate how requests are routed within this
    logical mesh boundary.

    Attributes:
        name (str):
            Required. Name of the Mesh resource. It matches pattern
            ``projects/*/locations/global/meshes/<mesh_name>``.
        self_link (str):
            Output only. Server-defined URL of this
            resource
        create_time (google.protobuf.timestamp_pb2.Timestamp):
            Output only. The timestamp when the resource
            was created.
        update_time (google.protobuf.timestamp_pb2.Timestamp):
            Output only. The timestamp when the resource
            was updated.
        labels (MutableMapping[str, str]):
            Optional. Set of label tags associated with
            the Mesh resource.
        description (str):
            Optional. A free-text description of the
            resource. Max length 1024 characters.
        interception_port (int):
            Optional. If set to a valid TCP port
            (1-65535), instructs the SIDECAR proxy to listen
            on the specified port of localhost (127.0.0.1)
            address. The SIDECAR proxy will expect all
            traffic to be redirected to this port regardless
            of its actual ip:port destination. If unset, a
            port '15001' is used as the interception port.
            This is applicable only for sidecar proxy
            deployments.
    """

    name: str = proto.Field(
        proto.STRING,
        number=1,
    )
    self_link: str = proto.Field(
        proto.STRING,
        number=9,
    )
    create_time: timestamp_pb2.Timestamp = proto.Field(
        proto.MESSAGE,
        number=2,
        message=timestamp_pb2.Timestamp,
    )
    update_time: timestamp_pb2.Timestamp = proto.Field(
        proto.MESSAGE,
        number=3,
        message=timestamp_pb2.Timestamp,
    )
    labels: MutableMapping[str, str] = proto.MapField(
        proto.STRING,
        proto.STRING,
        number=4,
    )
    description: str = proto.Field(
        proto.STRING,
        number=5,
    )
    interception_port: int = proto.Field(
        proto.INT32,
        number=8,
    )


class ServiceBinding(proto.Message):
    r"""ServiceBinding is the resource that defines a Service
    Directory Service to be used in a BackendService resource.

    Attributes:
        name (str):
            Required. Name of the ServiceBinding resource. It matches
            pattern
            ``projects/*/locations/global/serviceBindings/service_binding_name``.
        description (str):
            Optional. A free-text description of the
            resource. Max length 1024 characters.
        create_time (google.protobuf.timestamp_pb2.Timestamp):
            Output only. The timestamp when the resource
            was created.
        update_time (google.protobuf.timestamp_pb2.Timestamp):
            Output only. The timestamp when the resource
            was updated.
        service (str):
            Required. The full Service Directory Service name of the
            format projects/*/locations/*/namespaces/*/services/*
        service_id (str):
            Output only. The unique identifier of the
            Service Directory Service against which the
            Service Binding resource is validated. This is
            populated when the Service Binding resource is
            used in another resource (like Backend Service).
            This is of the UUID4 format.
        labels (MutableMapping[str, str]):
            Optional. Set of label tags associated with
            the ServiceBinding resource.
    """

    name: str = proto.Field(
        proto.STRING,
        number=1,
    )
    description: str = proto.Field(
        proto.STRING,
        number=2,
    )
    create_time: timestamp_pb2.Timestamp = proto.Field(
        proto.MESSAGE,
        number=3,
        message=timestamp_pb2.Timestamp,
    )
    update_time: timestamp_pb2.Timestamp = proto.Field(
        proto.MESSAGE,
        number=4,
        message=timestamp_pb2.Timestamp,
    )
    service: str = proto.Field(
        proto.STRING,
        number=5,
    )
    service_id: str = proto.Field(
        proto.STRING,
        number=8,
    )
    labels: MutableMapping[str, str] = proto.MapField(
        proto.STRING,
        proto.STRING,
        number=7,
    )


class TcpRoute(proto.Message):
    r"""TcpRoute is the resource defining how TCP traffic should be
    routed by a Mesh/Gateway resource.

    Attributes:
        name (str):
            Required. Name of the TcpRoute resource. It matches pattern
            ``projects/*/locations/global/tcpRoutes/tcp_route_name>``.
        self_link (str):
            Output only. Server-defined URL of this
            resource
        create_time (google.protobuf.timestamp_pb2.Timestamp):
            Output only. The timestamp when the resource
            was created.
        update_time (google.protobuf.timestamp_pb2.Timestamp):
            Output only. The timestamp when the resource
            was updated.
        description (str):
            Optional. A free-text description of the
            resource. Max length 1024 characters.
        rules (MutableSequence[google.events.cloud.networkservices_v1.types.TcpRoute.RouteRule]):
            Required. Rules that define how traffic is
            routed and handled. At least one RouteRule must
            be supplied. If there are multiple rules then
            the action taken will be the first rule to
            match.
        meshes (MutableSequence[str]):
            Optional. Meshes defines a list of meshes this TcpRoute is
            attached to, as one of the routing rules to route the
            requests served by the mesh.

            Each mesh reference should match the pattern:
            ``projects/*/locations/global/meshes/<mesh_name>``

            The attached Mesh should be of a type SIDECAR
        gateways (MutableSequence[str]):
            Optional. Gateways defines a list of gateways this TcpRoute
            is attached to, as one of the routing rules to route the
            requests served by the gateway.

            Each gateway reference should match the pattern:
            ``projects/*/locations/global/gateways/<gateway_name>``
        labels (MutableMapping[str, str]):
            Optional. Set of label tags associated with
            the TcpRoute resource.
    """

    class RouteRule(proto.Message):
        r"""Specifies how to match traffic and how to route traffic when
        traffic is matched.

        Attributes:
            matches (MutableSequence[google.events.cloud.networkservices_v1.types.TcpRoute.RouteMatch]):
                Optional. RouteMatch defines the predicate
                used to match requests to a given action.
                Multiple match types are "OR"ed for evaluation.
                If no routeMatch field is specified, this rule
                will unconditionally match traffic.
            action (google.events.cloud.networkservices_v1.types.TcpRoute.RouteAction):
                Required. The detailed rule defining how to
                route matched traffic.
        """

        matches: MutableSequence['TcpRoute.RouteMatch'] = proto.RepeatedField(
            proto.MESSAGE,
            number=1,
            message='TcpRoute.RouteMatch',
        )
        action: 'TcpRoute.RouteAction' = proto.Field(
            proto.MESSAGE,
            number=2,
            message='TcpRoute.RouteAction',
        )

    class RouteMatch(proto.Message):
        r"""RouteMatch defines the predicate used to match requests to a
        given action. Multiple match types are "OR"ed for evaluation. If
        no routeMatch field is specified, this rule will unconditionally
        match traffic.

        Attributes:
            address (str):
                Required. Must be specified in the CIDR range
                format. A CIDR range consists of an IP Address
                and a prefix length to construct the subnet
                mask. By default, the prefix length is 32 (i.e.
                matches a single IP address). Only IPV4
                addresses are supported. Examples: "10.0.0.1" -
                matches against this exact IP address.
                "10.0.0.0/8" - matches against any IP address
                within the 10.0.0.0 subnet and 255.255.255.0
                mask. "0.0.0.0/0"
                - matches against any IP address'.
            port (str):
                Required. Specifies the destination port to
                match against.
        """

        address: str = proto.Field(
            proto.STRING,
            number=1,
        )
        port: str = proto.Field(
            proto.STRING,
            number=2,
        )

    class RouteAction(proto.Message):
        r"""The specifications for routing traffic and applying
        associated policies.

        Attributes:
            destinations (MutableSequence[google.events.cloud.networkservices_v1.types.TcpRoute.RouteDestination]):
                Optional. The destination services to which
                traffic should be forwarded. At least one
                destination service is required. Only one of
                route destination or original destination can be
                set.
            original_destination (bool):
                Optional. If true, Router will use the
                destination IP and port of the original
                connection as the destination of the request.
                Default is false. Only one of route destinations
                or original destination can be set.
        """

        destinations: MutableSequence['TcpRoute.RouteDestination'] = proto.RepeatedField(
            proto.MESSAGE,
            number=1,
            message='TcpRoute.RouteDestination',
        )
        original_destination: bool = proto.Field(
            proto.BOOL,
            number=3,
        )

    class RouteDestination(proto.Message):
        r"""Describe the destination for traffic to be routed to.

        Attributes:
            service_name (str):
                Required. The URL of a BackendService to
                route traffic to.
            weight (int):
                Optional. Specifies the proportion of
                requests forwarded to the backend referenced by
                the serviceName field. This is computed as:

                        weight/Sum(weights in this destination
                list). For non-zero values, there may be some
                epsilon from the exact proportion defined here
                depending on the precision an implementation
                supports.

                If only one serviceName is specified and it has
                a weight greater than 0, 100% of the traffic is
                forwarded to that backend.

                If weights are specified for any one service
                name, they need to be specified for all of them.

                If weights are unspecified for all services,
                then, traffic is distributed in equal
                proportions to all of them.
        """

        service_name: str = proto.Field(
            proto.STRING,
            number=1,
        )
        weight: int = proto.Field(
            proto.INT32,
            number=2,
        )

    name: str = proto.Field(
        proto.STRING,
        number=1,
    )
    self_link: str = proto.Field(
        proto.STRING,
        number=11,
    )
    create_time: timestamp_pb2.Timestamp = proto.Field(
        proto.MESSAGE,
        number=2,
        message=timestamp_pb2.Timestamp,
    )
    update_time: timestamp_pb2.Timestamp = proto.Field(
        proto.MESSAGE,
        number=3,
        message=timestamp_pb2.Timestamp,
    )
    description: str = proto.Field(
        proto.STRING,
        number=4,
    )
    rules: MutableSequence[RouteRule] = proto.RepeatedField(
        proto.MESSAGE,
        number=5,
        message=RouteRule,
    )
    meshes: MutableSequence[str] = proto.RepeatedField(
        proto.STRING,
        number=8,
    )
    gateways: MutableSequence[str] = proto.RepeatedField(
        proto.STRING,
        number=9,
    )
    labels: MutableMapping[str, str] = proto.MapField(
        proto.STRING,
        proto.STRING,
        number=10,
    )


class TlsRoute(proto.Message):
    r"""TlsRoute defines how traffic should be routed based on SNI
    and other matching L3 attributes.

    Attributes:
        name (str):
            Required. Name of the TlsRoute resource. It matches pattern
            ``projects/*/locations/global/tlsRoutes/tls_route_name>``.
        self_link (str):
            Output only. Server-defined URL of this
            resource
        create_time (google.protobuf.timestamp_pb2.Timestamp):
            Output only. The timestamp when the resource
            was created.
        update_time (google.protobuf.timestamp_pb2.Timestamp):
            Output only. The timestamp when the resource
            was updated.
        description (str):
            Optional. A free-text description of the
            resource. Max length 1024 characters.
        rules (MutableSequence[google.events.cloud.networkservices_v1.types.TlsRoute.RouteRule]):
            Required. Rules that define how traffic is
            routed and handled. At least one RouteRule must
            be supplied. If there are multiple rules then
            the action taken will be the first rule to
            match.
        meshes (MutableSequence[str]):
            Optional. Meshes defines a list of meshes this TlsRoute is
            attached to, as one of the routing rules to route the
            requests served by the mesh.

            Each mesh reference should match the pattern:
            ``projects/*/locations/global/meshes/<mesh_name>``

            The attached Mesh should be of a type SIDECAR
        gateways (MutableSequence[str]):
            Optional. Gateways defines a list of gateways this TlsRoute
            is attached to, as one of the routing rules to route the
            requests served by the gateway.

            Each gateway reference should match the pattern:
            ``projects/*/locations/global/gateways/<gateway_name>``
    """

    class RouteRule(proto.Message):
        r"""Specifies how to match traffic and how to route traffic when
        traffic is matched.

        Attributes:
            matches (MutableSequence[google.events.cloud.networkservices_v1.types.TlsRoute.RouteMatch]):
                Required. RouteMatch defines the predicate
                used to match requests to a given action.
                Multiple match types are "OR"ed for evaluation.
            action (google.events.cloud.networkservices_v1.types.TlsRoute.RouteAction):
                Required. The detailed rule defining how to
                route matched traffic.
        """

        matches: MutableSequence['TlsRoute.RouteMatch'] = proto.RepeatedField(
            proto.MESSAGE,
            number=1,
            message='TlsRoute.RouteMatch',
        )
        action: 'TlsRoute.RouteAction' = proto.Field(
            proto.MESSAGE,
            number=2,
            message='TlsRoute.RouteAction',
        )

    class RouteMatch(proto.Message):
        r"""RouteMatch defines the predicate used to match requests to a
        given action. Multiple match types are "AND"ed for evaluation.
        If no routeMatch field is specified, this rule will
        unconditionally match traffic.

        Attributes:
            sni_host (MutableSequence[str]):
                Optional. SNI (server name indicator) to match against. SNI
                will be matched against all wildcard domains, i.e.
                ``www.example.com`` will be first matched against
                ``www.example.com``, then ``*.example.com``, then ``*.com.``
                Partial wildcards are not supported, and values like
                \*w.example.com are invalid. At least one of sni_host and
                alpn is required. Up to 5 sni hosts across all matches can
                be set.
            alpn (MutableSequence[str]):
                Optional. ALPN (Application-Layer Protocol Negotiation) to
                match against. Examples: "http/1.1", "h2". At least one of
                sni_host and alpn is required. Up to 5 alpns across all
                matches can be set.
        """

        sni_host: MutableSequence[str] = proto.RepeatedField(
            proto.STRING,
            number=1,
        )
        alpn: MutableSequence[str] = proto.RepeatedField(
            proto.STRING,
            number=2,
        )

    class RouteAction(proto.Message):
        r"""The specifications for routing traffic and applying
        associated policies.

        Attributes:
            destinations (MutableSequence[google.events.cloud.networkservices_v1.types.TlsRoute.RouteDestination]):
                Required. The destination services to which
                traffic should be forwarded. At least one
                destination service is required.
        """

        destinations: MutableSequence['TlsRoute.RouteDestination'] = proto.RepeatedField(
            proto.MESSAGE,
            number=1,
            message='TlsRoute.RouteDestination',
        )

    class RouteDestination(proto.Message):
        r"""Describe the destination for traffic to be routed to.

        Attributes:
            service_name (str):
                Required. The URL of a BackendService to
                route traffic to.
            weight (int):
                Optional. Specifies the proportion of requests forwareded to
                the backend referenced by the service_name field. This is
                computed as: weight/Sum(weights in destinations) Weights in
                all destinations does not need to sum up to 100.
        """

        service_name: str = proto.Field(
            proto.STRING,
            number=1,
        )
        weight: int = proto.Field(
            proto.INT32,
            number=2,
        )

    name: str = proto.Field(
        proto.STRING,
        number=1,
    )
    self_link: str = proto.Field(
        proto.STRING,
        number=8,
    )
    create_time: timestamp_pb2.Timestamp = proto.Field(
        proto.MESSAGE,
        number=2,
        message=timestamp_pb2.Timestamp,
    )
    update_time: timestamp_pb2.Timestamp = proto.Field(
        proto.MESSAGE,
        number=3,
        message=timestamp_pb2.Timestamp,
    )
    description: str = proto.Field(
        proto.STRING,
        number=4,
    )
    rules: MutableSequence[RouteRule] = proto.RepeatedField(
        proto.MESSAGE,
        number=5,
        message=RouteRule,
    )
    meshes: MutableSequence[str] = proto.RepeatedField(
        proto.STRING,
        number=6,
    )
    gateways: MutableSequence[str] = proto.RepeatedField(
        proto.STRING,
        number=7,
    )


class EndpointPolicyEventData(proto.Message):
    r"""The data within all EndpointPolicy events.

    .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

    Attributes:
        payload (google.events.cloud.networkservices_v1.types.EndpointPolicy):
            Optional. The EndpointPolicy event payload.
            Unset for deletion events.

            This field is a member of `oneof`_ ``_payload``.
    """

    payload: 'EndpointPolicy' = proto.Field(
        proto.MESSAGE,
        number=1,
        optional=True,
        message='EndpointPolicy',
    )


class HttpRouteEventData(proto.Message):
    r"""The data within all HttpRoute events.

    .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

    Attributes:
        payload (google.events.cloud.networkservices_v1.types.HttpRoute):
            Optional. The HttpRoute event payload. Unset
            for deletion events.

            This field is a member of `oneof`_ ``_payload``.
    """

    payload: 'HttpRoute' = proto.Field(
        proto.MESSAGE,
        number=1,
        optional=True,
        message='HttpRoute',
    )


class ServiceBindingEventData(proto.Message):
    r"""The data within all ServiceBinding events.

    .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

    Attributes:
        payload (google.events.cloud.networkservices_v1.types.ServiceBinding):
            Optional. The ServiceBinding event payload.
            Unset for deletion events.

            This field is a member of `oneof`_ ``_payload``.
    """

    payload: 'ServiceBinding' = proto.Field(
        proto.MESSAGE,
        number=1,
        optional=True,
        message='ServiceBinding',
    )


class GatewayEventData(proto.Message):
    r"""The data within all Gateway events.

    .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

    Attributes:
        payload (google.events.cloud.networkservices_v1.types.Gateway):
            Optional. The Gateway event payload. Unset
            for deletion events.

            This field is a member of `oneof`_ ``_payload``.
    """

    payload: 'Gateway' = proto.Field(
        proto.MESSAGE,
        number=1,
        optional=True,
        message='Gateway',
    )


class TlsRouteEventData(proto.Message):
    r"""The data within all TlsRoute events.

    .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

    Attributes:
        payload (google.events.cloud.networkservices_v1.types.TlsRoute):
            Optional. The TlsRoute event payload. Unset
            for deletion events.

            This field is a member of `oneof`_ ``_payload``.
    """

    payload: 'TlsRoute' = proto.Field(
        proto.MESSAGE,
        number=1,
        optional=True,
        message='TlsRoute',
    )


class GrpcRouteEventData(proto.Message):
    r"""The data within all GrpcRoute events.

    .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

    Attributes:
        payload (google.events.cloud.networkservices_v1.types.GrpcRoute):
            Optional. The GrpcRoute event payload. Unset
            for deletion events.

            This field is a member of `oneof`_ ``_payload``.
    """

    payload: 'GrpcRoute' = proto.Field(
        proto.MESSAGE,
        number=1,
        optional=True,
        message='GrpcRoute',
    )


class MeshEventData(proto.Message):
    r"""The data within all Mesh events.

    .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

    Attributes:
        payload (google.events.cloud.networkservices_v1.types.Mesh):
            Optional. The Mesh event payload. Unset for
            deletion events.

            This field is a member of `oneof`_ ``_payload``.
    """

    payload: 'Mesh' = proto.Field(
        proto.MESSAGE,
        number=1,
        optional=True,
        message='Mesh',
    )


class TcpRouteEventData(proto.Message):
    r"""The data within all TcpRoute events.

    .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

    Attributes:
        payload (google.events.cloud.networkservices_v1.types.TcpRoute):
            Optional. The TcpRoute event payload. Unset
            for deletion events.

            This field is a member of `oneof`_ ``_payload``.
    """

    payload: 'TcpRoute' = proto.Field(
        proto.MESSAGE,
        number=1,
        optional=True,
        message='TcpRoute',
    )


__all__ = tuple(sorted(__protobuf__.manifest))



# -*- coding: utf-8 -*-
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict
import os
import re
from typing import Dict, Optional, Type, Union

from google.api_core import client_options as client_options_lib  # type: ignore
from google.api_core import gapic_v1  # type: ignore
from google.api_core.operations_v1.transports.base import (
    DEFAULT_CLIENT_INFO,
    OperationsTransport,
)
from google.api_core.operations_v1.transports.rest import OperationsRestTransport

try:
    from google.api_core.operations_v1.transports.rest_asyncio import (
        AsyncOperationsRestTransport,
    )

    HAS_ASYNC_REST_DEPENDENCIES = True
except ImportError as e:
    HAS_ASYNC_REST_DEPENDENCIES = False
    ASYNC_REST_EXCEPTION = e

from google.auth import credentials as ga_credentials  # type: ignore
from google.auth.exceptions import MutualTLSChannelError  # type: ignore
from google.auth.transport import mtls  # type: ignore


class AbstractOperationsBaseClientMeta(type):
    """Metaclass for the Operations Base client.

    This provides base class-level methods for building and retrieving
    support objects (e.g. transport) without polluting the client instance
    objects.
    """

    _transport_registry = OrderedDict()  # type: Dict[str, Type[OperationsTransport]]
    _transport_registry["rest"] = OperationsRestTransport
    if HAS_ASYNC_REST_DEPENDENCIES:
        _transport_registry["rest_asyncio"] = AsyncOperationsRestTransport

    def get_transport_class(
        cls,
        label: Optional[str] = None,
    ) -> Type[OperationsTransport]:
        """Returns an appropriate transport class.

        Args:
            label: The name of the desired transport. If none is
                provided, then the first transport in the registry is used.

        Returns:
            The transport class to use.
        """
        # If a specific transport is requested, return that one.
        if (
            label == "rest_asyncio" and not HAS_ASYNC_REST_DEPENDENCIES
        ):  # pragma: NO COVER
            raise ASYNC_REST_EXCEPTION

        if label:
            return cls._transport_registry[label]

        # No transport is requested; return the default (that is, the first one
        # in the dictionary).
        return next(iter(cls._transport_registry.values()))


class AbstractOperationsBaseClient(metaclass=AbstractOperationsBaseClientMeta):
    """Manages long-running operations with an API service.

    When an API method normally takes long time to complete, it can be
    designed to return [Operation][google.api_core.operations_v1.Operation] to the
    client, and the client can use this interface to receive the real
    response asynchronously by polling the operation resource, or pass
    the operation resource to another API (such as Google Cloud Pub/Sub
    API) to receive the response. Any API service that returns
    long-running operations should implement the ``Operations``
    interface so developers can have a consistent client experience.
    """

    @staticmethod
    def _get_default_mtls_endpoint(api_endpoint):
        """Converts api endpoint to mTLS endpoint.

        Convert "*.sandbox.googleapis.com" and "*.googleapis.com" to
        "*.mtls.sandbox.googleapis.com" and "*.mtls.googleapis.com" respectively.
        Args:
            api_endpoint (Optional[str]): the api endpoint to convert.
        Returns:
            str: converted mTLS api endpoint.
        """
        if not api_endpoint:
            return api_endpoint

        mtls_endpoint_re = re.compile(
            r"(?P<name>[^.]+)(?P<mtls>\.mtls)?(?P<sandbox>\.sandbox)?(?P<googledomain>\.googleapis\.com)?"
        )

        m = mtls_endpoint_re.match(api_endpoint)
        name, mtls, sandbox, googledomain = m.groups()
        if mtls or not googledomain:
            return api_endpoint

        if sandbox:
            return api_endpoint.replace(
                "sandbox.googleapis.com", "mtls.sandbox.googleapis.com"
            )

        return api_endpoint.replace(".googleapis.com", ".mtls.googleapis.com")

    DEFAULT_ENDPOINT = "longrunning.googleapis.com"
    DEFAULT_MTLS_ENDPOINT = _get_default_mtls_endpoint.__func__(  # type: ignore
        DEFAULT_ENDPOINT
    )

    @classmethod
    def from_service_account_info(cls, info: dict, *args, **kwargs):
        """
        This class method should be overridden by the subclasses.

        Args:
            info (dict): The service account private key info.
            args: Additional arguments to pass to the constructor.
            kwargs: Additional arguments to pass to the constructor.

        Raises:
            NotImplementedError: If the method is called on the base class.
        """
        raise NotImplementedError("`from_service_account_info` is not implemented.")

    @classmethod
    def from_service_account_file(cls, filename: str, *args, **kwargs):
        """
        This class method should be overridden by the subclasses.

        Args:
            filename (str): The path to the service account private key json
                file.
            args: Additional arguments to pass to the constructor.
            kwargs: Additional arguments to pass to the constructor.

        Raises:
            NotImplementedError: If the method is called on the base class.
        """
        raise NotImplementedError("`from_service_account_file` is not implemented.")

    from_service_account_json = from_service_account_file

    @property
    def transport(self) -> OperationsTransport:
        """Returns the transport used by the client instance.

        Returns:
            OperationsTransport: The transport used by the client
                instance.
        """
        return self._transport

    @staticmethod
    def common_billing_account_path(
        billing_account: str,
    ) -> str:
        """Returns a fully-qualified billing_account string."""
        return "billingAccounts/{billing_account}".format(
            billing_account=billing_account,
        )

    @staticmethod
    def parse_common_billing_account_path(path: str) -> Dict[str, str]:
        """Parse a billing_account path into its component segments."""
        m = re.match(r"^billingAccounts/(?P<billing_account>.+?)$", path)
        return m.groupdict() if m else {}

    @staticmethod
    def common_folder_path(
        folder: str,
    ) -> str:
        """Returns a fully-qualified folder string."""
        return "folders/{folder}".format(
            folder=folder,
        )

    @staticmethod
    def parse_common_folder_path(path: str) -> Dict[str, str]:
        """Parse a folder path into its component segments."""
        m = re.match(r"^folders/(?P<folder>.+?)$", path)
        return m.groupdict() if m else {}

    @staticmethod
    def common_organization_path(
        organization: str,
    ) -> str:
        """Returns a fully-qualified organization string."""
        return "organizations/{organization}".format(
            organization=organization,
        )

    @staticmethod
    def parse_common_organization_path(path: str) -> Dict[str, str]:
        """Parse a organization path into its component segments."""
        m = re.match(r"^organizations/(?P<organization>.+?)$", path)
        return m.groupdict() if m else {}

    @staticmethod
    def common_project_path(
        project: str,
    ) -> str:
        """Returns a fully-qualified project string."""
        return "projects/{project}".format(
            project=project,
        )

    @staticmethod
    def parse_common_project_path(path: str) -> Dict[str, str]:
        """Parse a project path into its component segments."""
        m = re.match(r"^projects/(?P<project>.+?)$", path)
        return m.groupdict() if m else {}

    @staticmethod
    def common_location_path(
        project: str,
        location: str,
    ) -> str:
        """Returns a fully-qualified location string."""
        return "projects/{project}/locations/{location}".format(
            project=project,
            location=location,
        )

    @staticmethod
    def parse_common_location_path(path: str) -> Dict[str, str]:
        """Parse a location path into its component segments."""
        m = re.match(r"^projects/(?P<project>.+?)/locations/(?P<location>.+?)$", path)
        return m.groupdict() if m else {}

    def __init__(
        self,
        *,
        credentials: Optional[ga_credentials.Credentials] = None,
        transport: Union[str, OperationsTransport, None] = None,
        client_options: Optional[client_options_lib.ClientOptions] = None,
        client_info: gapic_v1.client_info.ClientInfo = DEFAULT_CLIENT_INFO,
    ) -> None:
        """Instantiates the operations client.

        Args:
            credentials (Optional[google.auth.credentials.Credentials]): The
                authorization credentials to attach to requests. These
                credentials identify the application to the service; if none
                are specified, the client will attempt to ascertain the
                credentials from the environment.
            transport (Union[str, OperationsTransport]): The
                transport to use. If set to None, a transport is chosen
                automatically.
            client_options (google.api_core.client_options.ClientOptions): Custom options for the
                client. It won't take effect if a ``transport`` instance is provided.
                (1) The ``api_endpoint`` property can be used to override the
                default endpoint provided by the client. GOOGLE_API_USE_MTLS_ENDPOINT
                environment variable can also be used to override the endpoint:
                "always" (always use the default mTLS endpoint), "never" (always
                use the default regular endpoint) and "auto" (auto switch to the
                default mTLS endpoint if client certificate is present, this is
                the default value). However, the ``api_endpoint`` property takes
                precedence if provided.
                (2) If GOOGLE_API_USE_CLIENT_CERTIFICATE environment variable
                is "true", then the ``client_cert_source`` property can be used
                to provide client certificate for mutual TLS transport. If
                not provided, the default SSL client certificate will be used if
                present. If GOOGLE_API_USE_CLIENT_CERTIFICATE is "false" or not
                set, no client certificate will be used.
            client_info (google.api_core.gapic_v1.client_info.ClientInfo):
                The client info used to send a user-agent string along with
                API requests. If ``None``, then default info will be used.
                Generally, you only need to set this if you're developing
                your own client library.

        Raises:
            google.auth.exceptions.MutualTLSChannelError: If mutual TLS transport
                creation failed for any reason.
        """
        if isinstance(client_options, dict):
            client_options = client_options_lib.from_dict(client_options)
        if client_options is None:
            client_options = client_options_lib.ClientOptions()

        # Create SSL credentials for mutual TLS if needed.
        use_client_cert = os.getenv(
            "GOOGLE_API_USE_CLIENT_CERTIFICATE", "false"
        ).lower()
        if use_client_cert not in ("true", "false"):
            raise ValueError(
                "Environment variable `GOOGLE_API_USE_CLIENT_CERTIFICATE` must be either `true` or `false`"
            )
        client_cert_source_func = None
        is_mtls = False
        if use_client_cert == "true":
            if client_options.client_cert_source:
                is_mtls = True
                client_cert_source_func = client_options.client_cert_source
            else:
                is_mtls = mtls.has_default_client_cert_source()
                if is_mtls:
                    client_cert_source_func = mtls.default_client_cert_source()
                else:
                    client_cert_source_func = None

        # Figure out which api endpoint to use.
        if client_options.api_endpoint is not None:
            api_endpoint = client_options.api_endpoint
        else:
            use_mtls_env = os.getenv("GOOGLE_API_USE_MTLS_ENDPOINT", "auto")
            if use_mtls_env == "never":
                api_endpoint = self.DEFAULT_ENDPOINT
            elif use_mtls_env == "always":
                api_endpoint = self.DEFAULT_MTLS_ENDPOINT
            elif use_mtls_env == "auto":
                if is_mtls:
                    api_endpoint = self.DEFAULT_MTLS_ENDPOINT
                else:
                    api_endpoint = self.DEFAULT_ENDPOINT
            else:
                raise MutualTLSChannelError(
                    "Unsupported GOOGLE_API_USE_MTLS_ENDPOINT value. Accepted "
                    "values: never, auto, always"
                )

        # Save or instantiate the transport.
        # Ordinarily, we provide the transport, but allowing a custom transport
        # instance provides an extensibility point for unusual situations.
        if isinstance(transport, OperationsTransport):
            # transport is a OperationsTransport instance.
            if credentials or client_options.credentials_file:
                raise ValueError(
                    "When providing a transport instance, "
                    "provide its credentials directly."
                )
            if client_options.scopes:
                raise ValueError(
                    "When providing a transport instance, provide its scopes "
                    "directly."
                )
            self._transport = transport
        else:
            Transport = type(self).get_transport_class(transport)
            self._transport = Transport(
                credentials=credentials,
                credentials_file=client_options.credentials_file,
                host=api_endpoint,
                scopes=client_options.scopes,
                client_cert_source_for_mtls=client_cert_source_func,
                quota_project_id=client_options.quota_project_id,
                client_info=client_info,
                always_use_jwt_access=True,
            )
